# -*- coding: utf-8 -*-
#
#
#  TheVirtualBrain-Scientific Package. This package holds all simulators, and 
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2017, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""

Tests all the possible combinations of (available) models and integration 
schemes (region and surface based simulations).

.. moduleauthor:: Paula Sanz Leon <sanzleon.paula@gmail.com@>
.. moduleauthor:: Stuart A. Knock <Stuart@tvb.invalid>

"""

import pytest
import numpy
import itertools
from tvb.tests.library.base_testcase import BaseTestCase
from tvb.simulator.common import get_logger
from tvb.simulator import simulator, models, coupling, integrators, monitors, noise
from tvb.datatypes.connectivity import Connectivity
from tvb.datatypes.cortex import Cortex
from tvb.datatypes.local_connectivity import LocalConnectivity
from tvb.datatypes.region_mapping import RegionMapping
from tvb.simulator.integrators import HeunDeterministic, IntegratorStochastic

LOG = get_logger(__name__)

MODEL_CLASSES = models.Model.get_known_subclasses().values()
METHOD_CLASSES = integrators.Integrator.get_known_subclasses().values()


class Simulator(object):
    """
    Simulator test class
    
    """

    def __init__(self):
        """
        Initialise the structural information, coupling function, and monitors.
        
        """

        # Initialise some Monitors with period in physical time
        raw = monitors.Raw()
        gavg = monitors.GlobalAverage(period=2 ** -2)
        subsamp = monitors.SubSample(period=2 ** -2)
        tavg = monitors.TemporalAverage(period=2 ** -2)
        eeg = monitors.EEG.from_file()
        eeg.period = 2 ** -2
        eeg2 = monitors.EEG.from_file()
        eeg2.period = 2 ** -2
        eeg2.reference = 'Fp2'  # EEG with a reference electrode
        meg = monitors.MEG.from_file()
        meg.period = 2 ** -2

        self.monitors = (raw, gavg, subsamp, tavg, eeg, eeg2, meg)

        self.method = None
        self.sim = None

    def run_simulation(self, simulation_length=2 ** 2):
        """
        Test a simulator constructed with one of the <model>_<scheme> methods.
        """

        results = [[] for _ in self.monitors]

        for step in self.sim(simulation_length=simulation_length):
            for i, result in enumerate(step):
                if result is not None:
                    results[i].append(result)

        return results

    def configure(self, dt=2 ** -3, model=models.Generic2dOscillator, speed=4.0,
                  coupling_strength=0.00042, method=HeunDeterministic,
                  surface_sim=False,
                  default_connectivity=True):
        """
        Create an instance of the Simulator class, by default use the
        generic plane oscillator local dynamic model and the deterministic 
        version of Heun's method for the numerical integration.
        
        """
        self.method = method

        if default_connectivity:
            white_matter = Connectivity.from_file()
            region_mapping = RegionMapping.from_file(source_file="regionMapping_16k_76.txt")
        else:
            white_matter = Connectivity.from_file(source_file="connectivity_192.zip")
            region_mapping = RegionMapping.from_file(source_file="regionMapping_16k_192.txt")

        white_matter_coupling = coupling.Linear(a=numpy.array([coupling_strength]))
        white_matter.speed = numpy.array([speed])  # no longer allow scalars to numpy array promotion

        dynamics = model()

        if issubclass(method, IntegratorStochastic):
            hisss = noise.Additive(nsig=numpy.array([2 ** -11]))
            integrator = method(dt=dt, noise=hisss)
        else:
            integrator = method(dt=dt)

        if surface_sim:
            local_coupling_strength = numpy.array([2 ** -10])
            default_cortex = Cortex.from_file()
            default_cortex.region_mapping_data = region_mapping
            default_cortex.coupling_strength = local_coupling_strength
            if default_connectivity:
                default_cortex.local_connectivity = LocalConnectivity.from_file()
            else:
                default_cortex.local_connectivity = LocalConnectivity()
            default_cortex.local_connectivity.surface = default_cortex
        else:
            default_cortex = None

        # Order of monitors determines order of returned values.
        self.sim = simulator.Simulator()
        self.sim.surface = default_cortex
        self.sim.model = dynamics
        self.sim.integrator = integrator
        self.sim.connectivity = white_matter
        self.sim.coupling = white_matter_coupling
        self.sim.monitors = self.monitors
        self.sim.configure()


class TestSimulator(BaseTestCase):
    @pytest.mark.slow
    @pytest.mark.parametrize('model_class,method_class', itertools.product(MODEL_CLASSES, METHOD_CLASSES))
    def test_simulator_region(self, model_class, method_class):
        test_simulator = Simulator()
        test_simulator.configure(model=model_class, method=method_class, surface_sim=False)
        result = test_simulator.run_simulation()

        self.assert_equal(len(test_simulator.monitors), len(result))
        for ts in result:
            assert ts is not None
            assert len(ts) > 0

    @pytest.mark.slow
    @pytest.mark.parametrize('default_connectivity', [True, False])
    def test_simulator_surface(self, default_connectivity):
        """
        This test evaluates if surface simulations run as basic flow.
        """
        test_simulator = Simulator()

        test_simulator.configure(surface_sim=True, default_connectivity=default_connectivity)
        result = test_simulator.run_simulation(simulation_length=2)

        assert len(test_simulator.monitors) == len(result)
        LOG.debug("Surface simulation finished for defaultConnectivity= %s" % str(default_connectivity))



# -*- coding: utf-8 -*-
#
#
#  TheVirtualBrain-Scientific Package. This package holds all simulators, and 
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2017, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""

The LookUpTable datatype. This brings together the scientific and framework 
methods that are associated with the precalculated look up tables.

.. moduleauthor:: Paula Sanz Leon <Paula@tvb.invalid>

"""

import numpy
from tvb.basic.readers import try_get_absolute_path
from tvb.basic.logger.builder import get_logger
from tvb.basic.neotraits.api import HasTraits, Attr, NArray, Int


LOG = get_logger(__name__)


class LookUpTable(HasTraits):
    """
    Lookup Tables for storing pre-computed functions.
    Specific table subclasses are implemented below.
    """

    _base_classes = ['LookUpTables']

    equation = Attr(
        field_type=str,
        label="String representation of the precalculated function",
        doc="""A latex representation of the function whose values are stored
                in the table, with the extra escaping needed for interpretation via sphinx.""")

    xmin = NArray(
        label="x-min",
        doc="""Minimum value""")

    xmax = NArray(
        label="x-max",
        doc="""Maximum value""")

    data = NArray(
        label="data",
        doc="""Tabulated values""")

    number_of_values = Int(
        label="Number of values",
        default=0,
        doc="""The number of values in the table """)

    df = NArray(
        label="df",
        doc=""".""")

    dx = NArray(
        label="dx",
        default=numpy.array([]),
        doc="""Tabulation step""")

    invdx = NArray(
        label="invdx",
        default=numpy.array([]),
        doc=""".""")

    @staticmethod
    def populate_table(result, source_file):
        source_full_path = try_get_absolute_path("tvb_data.tables", source_file)
        zip_data = numpy.load(source_full_path)

        result.df = zip_data['df']
        result.xmin, result.xmax = zip_data['min_max']
        result.data = zip_data['f']
        return result

    def configure(self):
        """
        Invoke the compute methods for computable attributes that haven't been
        set during initialization.
        """
        super(LookUpTable, self).configure()

        # Check if dx and invdx have been computed
        if self.number_of_values == 0:
            self.number_of_values = self.data.shape[0]

        if self.dx.size == 0:
            self.compute_search_indices()

    def summary_info(self):
        """
        Gather scientifically interesting summary information from an instance
        of this dataType, if any ...
        """
        return {"Number of values": self.number_of_values}

    def compute_search_indices(self):
        """
        ...
        """
        self.dx = ((self.xmax - self.xmin) / (self.number_of_values) - 1)
        self.invdx = 1 / self.dx

    def search_value(self, val):
        """
        Search a value in this look up table
        """

        if self.xmin:
            y = val - self.xmin
        else:
            y = val

        ind = numpy.array(y * self.invdx, dtype=int)

        try:
            return self.data[ind] + self.df[ind] * (y - ind * self.dx)
        except IndexError:  # out of bounds
            return numpy.NaN
            # NOTE: not sure if we should return a NaN or make val = self.max
            # At the moment, we force the input values to be within a known range


class PsiTable(LookUpTable):
    """
    Look up table containing the values of a function representing the time-averaged gating variable
    :math:`\\psi(\\nu)` as a function of the presynaptic rates :math:`\\nu`

    """

    @staticmethod
    def from_file(source_file="psi.npz", instance=None):
        return LookUpTable.populate_table(instance or PsiTable(), source_file)


class NerfTable(LookUpTable):
    """
    Look up table containing the values of Nerf integral within the :math:`\\phi`
    function that describes how the discharge rate vary as a function of parameters
    defining the statistical properties of the membrane potential in presence of synaptic inputs.

    """

    @staticmethod
    def from_file(source_file="nerf_int.npz", instance=None):
        return LookUpTable.populate_table(instance or NerfTable(), source_file)


# -*- coding: utf-8 -*-
#
#
#  TheVirtualBrain-Scientific Package. This package holds all simulators, and
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2017, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
Experimental OpenCL models implementations.

.. moduleauthor:: Marmaduke Woodman <mmwoodman@gmail.com>

"""

import pyopencl
import pyopencl.array
import numpy
from ..models import ReducedWongWang
DEBUG = True
class CLComponent(object):

    def configure_opencl(self, context, queue):
        self._context = context
        self._queue = queue
        if hasattr(self, '_opencl_program_source'):
            self._program = pyopencl.Program(context, self._opencl_program_source).build()
        elif hasattr(self, '_opencl_program_source_file'):
            self._program = pyopencl.Program(context, open(getattr(self,'_opencl_program_source_file'),'r').read()).build()
class CLModel(CLComponent):

    def _alloc_opencl(self, n_nodes ,n_states=1,n_mode=1):

        if not hasattr(self, '_queue'):
            msg = "OpenCL components must be configured via the `configure_opencl` method prior to use."
            raise RuntimeError(msg)

        # arrays in component workspace
        arrays = {'state': (n_states, n_nodes,n_mode), 'coupling': (1, n_nodes), 'deriv': (n_states, n_nodes, n_mode)}
        self.arrays = arrays
        if hasattr(self, '_opencl_ordered_params'):
            arrays['param'] = (len(self._opencl_ordered_params), n_nodes)
            #arrays['param'] = (len(self._opencl_ordered_params), n_nodes, n_mode)
        # alloc opencl device arrays
        self._arrays = {}
        for name, shape in arrays.items():
            self._arrays[name] = pyopencl.array.Array(self._queue, shape, 'f')

        # fill parameter values
        if hasattr(self, '_opencl_ordered_params'):
            if(DEBUG): print(self._opencl_ordered_params)
            for i, name in enumerate(self._opencl_ordered_params):
                if(DEBUG): print((i, name))
                val = getattr(self, name)

                if val.size == 1:
                    val = val.flat[0]

                self._arrays['param'][i] = val

        # setup kernel arguments
        self._kernel.set_args(*[self._arrays[key].data for key in 'state coupling param deriv'.split()])

    def configure_opencl(self, context, queue):
        super(CLModel, self).configure_opencl(context, queue)


        self._kernel = self._program.dfun

    def dfun(self , queue, nt, args ):
        self._kernel(queue, (int(nt),), None, *[arg.data for arg in args])


    def dfunKernel(self, state_variables, coupling, local_coupling=0.0 ):
        n_states = state_variables.shape[0]
        n_nodes = state_variables.shape[1]
        n_mode = state_variables.shape[2]
        # allocate data if not yet done so
        if not hasattr(self, '_arrays'):
            self._alloc_opencl(n_nodes,n_states = n_states,n_mode = n_mode)

        # copy if passed host arrays
        if isinstance(state_variables, numpy.ndarray):
            # state_variables, coupling will be (1, n, 1)
            if(DEBUG):
                print("state_variables are ndarray", "states:", state_variables.shape, "coupling:", coupling.shape)

            #self._arrays['state'][:] = state_variables.reshape((1, n_states*n_nodes*n_mode)).astype('f')
            #self._arrays['coupling'][:] = coupling.reshape((1, n_nodes)).astype('f')

            # self._arrays['state'] = state_variables.flatten()
            #self._arrays['coupling'] = coupling.reshape((1, n_nodes)).astype('f')
            if (DEBUG):
                print("state_variable shape:",
                      state_variables.reshape((n_states, n_nodes * n_mode, 1)).astype('f').shape)
                print("array state shape", self._arrays['state'][:].shape)
            self._arrays['state'][:] = state_variables.reshape((n_states,  n_nodes , n_mode)).astype('f')
            self._arrays['coupling'][:] = coupling.reshape((1, n_nodes)).astype('f')




        # set kernel arg if passed device arrays
        elif isinstance(state_variables, pyopencl.array.Array):
            self._kernel.set_args(state_variables.data,
                                  coupling.data,
                                  self._arrays['param'].data,
                                  self._arrays['deriv'].data)

        # otherwise, complain
        else:
            raise TypeError('unsupported data type %r', type(state_variables))

        # run the kernel and wait
        print("Run kernel...")

        pyopencl.enqueue_nd_range_kernel(self._queue, self._kernel, (n_nodes,), None).wait()

        # return derivatives following input type
        deriv = self._arrays['deriv']
        if(DEBUG):
            print("derive shape:", deriv.shape)
        if isinstance(state_variables, numpy.ndarray):
            deriv = deriv.get().reshape((n_states, n_nodes, n_mode)).astype('d')

        return deriv


class CLRWW(ReducedWongWang, CLModel):

    _opencl_ordered_params = 'a b d gamma tau_s w J_N I_o'.split()

    _opencl_program_source = """
    __kernel void dfun(__global float *state, __global float *coupling,
                       __global float *param, __global float *deriv)
    {
        int i = get_global_id(0), n = get_global_size(0);

        // this is boilerplate and could be generated
        float S=state[i], a=param[i], b=param[n+i], d=param[2*n+i], g=param[3*n+i],
              ts=param[4*n+i], w=param[5*n+i], j=param[6*n+i], io=param[7*n+i];

        float x = w*j*S + io + j*coupling[i];
        float h = (a*x - b) / (1.0f - exp(-d*(a*x - b)));
        float dx = - (S / ts) + (1.0f - S) * h * g;

        bool neg = S < 0.0f;
        bool gt1 = S > 1.0f;
        bool i01 = (!neg) && (!gt1);

        deriv[i] = neg*(0.0f - S) + gt1*(1.0f - S) + i01*dx;
    }
    """



# -*- coding: utf-8 -*-
#
#
#  TheVirtualBrain-Scientific Package. This package holds all simulators, and 
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2017, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
Helper tools, for the configuration area.

.. moduleauthor:: Bogdan Neacsa <bogdan.neacsa@codemart.ro>
.. moduleauthor:: marmaduke <duke@eml.cc>

"""


class EnhancedDictionary(dict):
    """
    A dictionary like class that provides easy access to configuration values.
    """

    def __getattr__(self, key):
        return self[key]

    def __setattr__(self, key, value):
        self[key] = value



class LibraryImportError(ImportError):
    """
    This is just a flag telling us at "reload" that a module was rejected due to our custom exception
    """



class LibraryModulesFinder():
    """
    In case users run TVB in 'library' profile access should be restricted to some parts of tvb,
    to avoid errors from those parts (which are not excepted to run with library settings).
    """

    restricted_modules = ['tvb.interfaces',
                          'tvb.datatype_removers',
                          'tvb.core',
                          'tvb.config',
                          'tvb.adapters']


    def find_module(self, fullname, path=None):

        for restricted in self.restricted_modules:
            if fullname.startswith(restricted):
                return self


    def load_module(self, module_name):
        info_message = str("You are trying to import the module `%s` in library mode."
                           "The library profile is a lightweight version of TVB and you "
                           "only have access to the simulator, analyzers and datatypes packages."
                           "If you want to use the entire TVB Framework start it either in command "
                           "or web interface profile." % module_name)
        raise LibraryImportError(info_message)




# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and 
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2017, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
# CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

import os
import collections
import numpy
import scipy.sparse
from tvb.basic.readers import try_get_absolute_path, FileReader
from tvb.basic.logger.builder import get_logger
from . import local_connectivity, region_mapping, surfaces
from tvb.basic.neotraits.api import Attr, NArray, Range, narray_describe

LOG = get_logger(__name__)


class Cortex(surfaces.CorticalSurface):
    """
    Wrapper Class over a CorticalSurface, to be used when preparing a simulation launch.
    """

    _ui_complex_datatype = surfaces.CorticalSurface

    _ui_name = "A cortex..."

    local_connectivity = Attr(
        field_type=local_connectivity.LocalConnectivity,
        label="Local Connectivity",
        required=False,
        doc="Define the interaction between neighboring network nodes. This is implicitly integrated in"
            " the definition of a given surface as an excitatory mean coupling of directly adjacent neighbors to"
            " the first state variable of each population model (since these typically represent the mean-neural"
            " membrane voltage). This coupling is instantaneous (no time delays).")

    region_mapping_data = Attr(
        field_type=region_mapping.RegionMapping,
        label="region mapping",
        doc="""An index vector of length equal to the number_of_vertices + the
            number of non-cortical regions, with values that index into an
            associated connectivity matrix.""")  # 'CS'

    region_areas = None
    region_orientation = None

    coupling_strength = NArray(
        label="Local coupling strength",
        domain=Range(lo=0.0, hi=20.0, step=1.0),
        default=numpy.array([1.0]),
        # file_storage=core.FILE_STORAGE_NONE,
        doc="""A factor that rescales local connectivity strengths.""")

    eeg_projection = NArray(
        label="EEG projection",
        # NOTE: This is redundant if the EEG monitor isn't used, but it makes life simpler.
        required=False,
        doc="""A 2-D array which projects the neural activity on the cortical
                surface to a set of EEG sensors."""
    )

    meg_projection = NArray(
        label="MEG projection",
        # linked = ?sensors, skull, skin, etc?
        doc="""A 2-D array which projects the neural activity on the cortical
            surface to a set of MEG sensors.""",
        required=False, )
    #  requires linked SensorsMEG

    internal_projection = NArray(
        label="Internal projection",
        required=False,
        doc="""A 2-D array which projects the neural activity on the
                cortical surface to a set of embeded sensors."""
    )
    #  requires linked SensorsInternal

    def populate_cortex(self, cortex_surface, cortex_parameters=None):
        """
        Populate 'self' from a CorticalSurfaceData instance with additional
        CortexData specific attributes.

        :param cortex_surface:  CorticalSurfaceData instance
        :param cortex_parameters: dictionary key:value, where key is attribute on CortexData
        """
        for name in cortex_surface.trait:  ##### todo: !!!!!!!!!!!!!
            try:
                setattr(self, name, getattr(cortex_surface, name))
            except Exception as exc:
                LOG.exception("Could not set attribute '" + name + "' on Cortex")
        for key, value in cortex_parameters.items():
            setattr(self, key, value)
        return self

    @property
    def region_mapping(self):
        """
        Define shortcut for retrieving RegionMapping map array.
        """
        if self.region_mapping_data is None:
            return None
        return self.region_mapping_data.array_data


    def configure(self):
        """
        Invoke the compute methods for computable attributes that haven't been
        set during initialisation.
        """
        super(Cortex, self).configure()

        if self.region_orientation is None:
            self.compute_region_orientation()

        if self.region_areas is None:
            self.compute_region_areas()

        if self.local_connectivity is None:
            self.local_connectivity = local_connectivity.LocalConnectivity(cutoff=40.0, surface=self)

        # mhtodo: review nullability of NArrays
        if self.local_connectivity.matrix is None or self.local_connectivity.matrix.size == 0:
            self.compute_local_connectivity()

        # Pad the local connectivity matrix with zeros when non-cortical regions
        # are included in the long range connectivity...
        if self.local_connectivity.matrix.shape[0] < self.region_mapping.shape[0]:
            LOG.info("There are non-cortical regions, will pad local connectivity")
            padding = scipy.sparse.csc_matrix((self.local_connectivity.matrix.shape[0],
                                               self.region_mapping.shape[0] - self.local_connectivity.matrix.shape[0]))
            self.local_connectivity.matrix = scipy.sparse.hstack([self.local_connectivity.matrix, padding])

            padding = scipy.sparse.csc_matrix((self.region_mapping.shape[0] - self.local_connectivity.matrix.shape[0],
                                               self.local_connectivity.matrix.shape[1]))
            self.local_connectivity.matrix = scipy.sparse.vstack([self.local_connectivity.matrix, padding])

    def summary_info(self):
        """
        Extend the base class's scientific summary information dictionary.
        """
        summary = super(Cortex, self).summary_info()
        summary["Number of regions"] = numpy.sum(self.region_areas > 0.0)
        summary["Region area, mean (mm:math:`^2`)"] = self.region_areas.mean()
        summary["Region area, minimum (mm:math:`^2`)"] = self.region_areas.min()
        summary["Region area, maximum (mm:math:`^2`)"] = self.region_areas.max()

        return summary

    def compute_local_connectivity(self):
        """
        """
        LOG.info("Computing local connectivity matrix")
        loc_con_cutoff = self.local_connectivity.cutoff
        self.compute_geodesic_distance_matrix(max_dist=loc_con_cutoff)

        self.local_connectivity.matrix_gdist = self.geodesic_distance_matrix.copy()
        self.local_connectivity.compute()  # Evaluate equation based distance

        #HACK FOR DEBUGGING CAUSE TRAITS REPORTS self.local_connectivity.trait["matrix"] AS BEING EMPTY...
        lcmat = self.local_connectivity.matrix
        sts = str(lcmat.__class__)
        # mhtodo: the trait.name is the file name for mapped,
        name = self.__class__.__name__ + ".local_connectivity"  # self.local_connectivity.trait.name))
        shape = str(lcmat.shape)
        sparse_format = str(lcmat.format)
        nnz = str(lcmat.nnz)
        dtype = str(lcmat.dtype)
        if lcmat.data.any() and lcmat.data.size > 0:
            array_max = lcmat.data.max()
            array_min = lcmat.data.min()
        else:
            array_max = array_min = 0.0
        LOG.debug("%s: %s shape: %s" % (sts, name, shape))
        LOG.debug("%s: %s format: %s" % (sts, name, sparse_format))
        LOG.debug("%s: %s number of non-zeros: %s" % (sts, name, nnz))
        LOG.debug("%s: %s dtype: %s" % (sts, name, dtype))
        LOG.debug("%s: %s maximum: %s" % (sts, name, array_max))
        LOG.debug("%s: %s minimum: %s" % (sts, name, array_min))

    def compute_region_areas(self):
        """Update the region_area attribute."""
        regions = numpy.unique(self.region_mapping)
        number_of_regions = len(regions)
        region_surface_area = numpy.zeros((number_of_regions, 1))
        avt = numpy.array(self.vertex_triangles)
        #NOTE: Slightly overestimates as it counts overlapping border triangles,
        #      but, not really a problem provided triangle-size << region-size.

        #NOTE: Check if there are non-cortical regions.

        if len(self.region_mapping) > len(self.vertex_normals):
            vertices_per_region = numpy.bincount(self.region_mapping)
            # Assume non-cortical regions will have len 1.
            non_cortical_regions, = numpy.where(vertices_per_region == 1)
            cortical_regions, = numpy.where(vertices_per_region > 1)
            #Average orientation of the region
            cortical_region_mapping = [x for x in self.region_mapping if x in cortical_regions]

            for nk in non_cortical_regions:
                region_surface_area[nk, :] = 0.0
            for k in cortical_regions:
                regs = list(map(set, avt[cortical_region_mapping == k]))
                region_triangles = set.union(*regs)
                region_surface_area[k] = self.triangle_areas[list(region_triangles)].sum()
        else:
            for k in regions:
                regs = list(map(set, avt[self.region_mapping == k]))
                region_triangles = set.union(*regs)
                region_surface_area[k] = self.triangle_areas[list(region_triangles)].sum()

        LOG.debug("region_areas")
        LOG.debug(narray_describe(region_surface_area))

        self.region_areas = region_surface_area

    def compute_region_orientation(self):
        """Update the region_orientation attribute."""
        regions = numpy.unique(self.region_mapping)
        average_orientation = numpy.zeros((len(regions), 3))
        if len(self.region_mapping) > len(self.vertex_normals):
            # Count how many vertices each region has.
            counter = collections.Counter(self.region_mapping)
            # Presumably non-cortical regions will have len 1 vertex assigned.
            vertices_per_region = numpy.asarray(list(dict(sorted(counter.items())).values()))
            non_cortical_regions = numpy.where(vertices_per_region == 1)
            cortical_regions = numpy.where(vertices_per_region > 1)
            cortical_region_mapping = [x for x in self.region_mapping if x in cortical_regions[0]]
            #Average orientation of the region

            for k in cortical_regions[0]:
                orient = self.vertex_normals[cortical_region_mapping == k, :]
                avg_orient = numpy.mean(orient, axis=0)
                average_orientation[k, :] = avg_orient / numpy.sqrt(numpy.sum(avg_orient ** 2))
            for nk in non_cortical_regions[0]:
                average_orientation[nk, :] = numpy.zeros((1, 3))
        else:
            #Average orientation of the region
            for k in regions:
                orient = self.vertex_normals[self.region_mapping == k, :]
                avg_orient = numpy.mean(orient, axis=0)
                average_orientation[k, :] = avg_orient / numpy.sqrt(numpy.sum(avg_orient ** 2))

        LOG.debug("region_orientation")
        LOG.debug(narray_describe(average_orientation))

        self.region_orientation = average_orientation

    @classmethod
    def from_file(cls, source_file="cortex_16384.zip",
                  region_mapping_file=os.path.join("regionMapping_16k_76.txt"),
                  local_connectivity_file=None, eeg_projection_file=None, instance=None):

        result = super(Cortex, cls).from_file(source_file, instance)

        if instance is not None:
            # Called through constructor directly
            if result.region_mapping is None:
                result.region_mapping_data = region_mapping.RegionMapping.from_file()

            if not result.eeg_projection:
                result.eeg_projection = Cortex.from_file_projection_array()

            if result.local_connectivity is None:
                result.local_connectivity = local_connectivity.LocalConnectivity.from_file()

        if region_mapping_file is not None:
            result.region_mapping_data = region_mapping.RegionMapping.from_file(region_mapping_file)

        if local_connectivity_file is not None:
            result.local_connectivity = local_connectivity.LocalConnectivity.from_file(local_connectivity_file)

        if eeg_projection_file is not None:
            result.eeg_projection = Cortex.from_file_projection_array(eeg_projection_file)

        return result

    @staticmethod
    def from_file_projection_array(source_file="projection_eeg_62_surface_16k.mat",
                                   matlab_data_name="ProjectionMatrix"):

        source_full_path = try_get_absolute_path("tvb_data.projectionMatrix", source_file)
        reader = FileReader(source_full_path)

        return reader.read_array(matlab_data_name=matlab_data_name)



# -*- coding: utf-8 -*-
#
#
#  TheVirtualBrain-Scientific Package. This package holds all simulators, and 
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2017, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
A collection of plotting functions used by tvb demos to plot 3D
Mayavi library is necessary in order for this module to work.

.. moduleauthor:: Stuart A. Knock <Stuart@tvb.invalid>
.. moduleauthor:: Paula Sanz Leon <paula.sanz-leon@univ-amu.fr>
"""

import numpy
import networkx as nx
from tvb.basic.logger.builder import get_logger

LOG = get_logger(__name__)

try:
    from mayavi import mlab


    @mlab.animate(delay=41, ui=True)
    def surface_timeseries(surface, data, step=1):
        """
        
        """
        fig = mlab.figure(figure="surface_timeseries", fgcolor=(0.5, 0.5, 0.5))
        # Plot an initial surface and colourbar
        surf_mesh = mlab.triangular_mesh(surface.vertices[:, 0],
                                         surface.vertices[:, 1],
                                         surface.vertices[:, 2],
                                         surface.triangles,
                                         scalars=data[0, :],
                                         vmin=data.min(), vmax=data.max(),
                                         figure=fig)
        mlab.colorbar(object=surf_mesh, orientation="vertical")

        # Handle for the surface object and figure
        surf = surf_mesh.mlab_source

        # Time
        tpts = data.shape[0]
        time_step = mlab.text(0.85, 0.125, ("0 of %s" % str(tpts)),
                              width=0.0625, color=(1, 1, 1), figure=fig,
                              name="counter")

        # Movie
        k = 0
        while 1:
            if abs(k) >= tpts:
                k = 0
            surf.set(scalars=data[k, :])
            time_step.set(text=("%s of %s" % (str(k), str(tpts))))
            k += step
            yield
        # mlab.show()


    def plot_surface(surface, fig=None, name=None, op=1.0, rep='surface'):
        """
        """
        if fig is None:
            fig = mlab.figure(figure=name, fgcolor=(0.5, 0.5, 0.5))

        surf_mesh = mlab.triangular_mesh(surface.vertices[:, 0],
                                         surface.vertices[:, 1],
                                         surface.vertices[:, 2],
                                         surface.triangles,
                                         color=(0.7, 0.67, 0.67),
                                         opacity=op,
                                         representation=rep,
                                         figure=fig)

        return surf_mesh


    def surface_orientation(surface, normals="triangles", name=None):
        """
        """
        fig = mlab.figure(figure=name, fgcolor=(0.5, 0.5, 0.5))
        surf_mesh = mlab.triangular_mesh(surface.vertices[:, 0],
                                         surface.vertices[:, 1],
                                         surface.vertices[:, 2],
                                         surface.triangles,
                                         color=(0.7, 0.67, 0.67),
                                         figure=fig)
        surf_orient = None
        if normals == "triangles":
            surf_orient = mlab.quiver3d(surface.triangle_centres[:, 0],
                                        surface.triangle_centres[:, 1],
                                        surface.triangle_centres[:, 2],
                                        surface.triangle_normals[:, 0],
                                        surface.triangle_normals[:, 1],
                                        surface.triangle_normals[:, 2])
        elif normals == "vertices":
            surf_orient = mlab.quiver3d(surface.vertices[:, 0],
                                        surface.vertices[:, 1],
                                        surface.vertices[:, 2],
                                        surface.vertex_normals[:, 0],
                                        surface.vertex_normals[:, 1],
                                        surface.vertex_normals[:, 2])
        else:
            LOG.error("normals must be either 'triangles' or 'vertices'")

        return surf_mesh, surf_orient


    def surface_parcellation(cortex_boundaries, colouring, mapping_colours, colour_rgb, interaction=False):
        """
        """
        number_of_vertices = cortex_boundaries.cortex.vertices.shape[0]
        number_of_triangles = cortex_boundaries.cortex.triangles.shape[0]

        number_of_regions = len(cortex_boundaries.region_neighbours)
        alpha = 255
        lut = numpy.zeros((number_of_regions, 4), dtype=numpy.uint8)
        for k in range(number_of_regions):
            lut[k] = numpy.hstack((colour_rgb[mapping_colours[colouring[k]]], alpha))

        fig = mlab.figure(figure="surface parcellation", bgcolor=(0.0, 0.0, 0.0), fgcolor=(0.5, 0.5, 0.5))
        surf_mesh = mlab.triangular_mesh(cortex_boundaries.cortex.vertices[:number_of_vertices // 2, 0],
                                         cortex_boundaries.cortex.vertices[:number_of_vertices // 2, 1],
                                         cortex_boundaries.cortex.vertices[:number_of_vertices // 2, 2],
                                         cortex_boundaries.cortex.triangles[:number_of_triangles // 2, :],
                                         scalars=cortex_boundaries.cortex.region_mapping[:number_of_vertices // 2],
                                         figure=fig)
        surf_mesh.module_manager.scalar_lut_manager.lut.number_of_colors = number_of_regions

        x = cortex_boundaries.boundary[:, 0]
        y = cortex_boundaries.boundary[:, 1]
        z = cortex_boundaries.boundary[:, 2]
        bpts = mlab.points3d(x, y, z, color=(0.25, 0.25, 0.25), scale_factor=1)
        mlab.show(stop=interaction)
        return surf_mesh, bpts


    def surface_pattern(surface, vertex_colours, custom_lut=None, foci=None):
        """
        Plot a surface and colour it based on a vector of length number of 
        vertices (vertex_colours).

        * How to obtain a pretty picture (from Mayavi's gui): 
        
          - set surf_mesh color to rgb(237, 217, 221)
          - add a surface module derived from surf_mesh; set 'Actor' 
            representation to wireframe; colour 'gray'.
          - enable contours of scalar_surf  
        """
        fig = mlab.figure(figure="surface pattern", fgcolor=(0.5, 0.5, 0.5))
        surf_mesh = mlab.triangular_mesh(surface.vertices[:, 0],
                                         surface.vertices[:, 1],
                                         surface.vertices[:, 2],
                                         surface.triangles,
                                         figure=fig)
        sm_obj = surf_mesh.mlab_source
        scalar_data = surf_mesh.mlab_source.dataset.point_data
        scalar_data.scalars = vertex_colours
        scalar_data.scalars.name = 'Scalar data'
        scalar_data.update()
        scalar_mesh = mlab.pipeline.set_active_attribute(surf_mesh, point_scalars='Scalar data')
        scalar_surf = mlab.pipeline.surface(scalar_mesh)

        if custom_lut is not None:
            # and finally we put this LUT back in the surface object. We could have
            # added any 255*4 array rather than modifying an existing LUT.
            scalar_surf.module_manager.scalar_lut_manager.lut.table = custom_lut

        if foci is not None:
            mlab.points3d(foci[:, 0],
                          foci[:, 1],
                          foci[:, 2],
                          scale_factor=2.,
                          scale_mode='none',
                          resolution=5,
                          opacity=0.01)

        mlab.show(stop=True)
        return sm_obj


    def xmas_balls(connectivity,
                   labels=True, labels_indices=None,
                   balls_colormap='Blues',
                   bgcolor=(0.5, 0.5, 0.5),
                   node_data=None, node_size=4.,
                   edge_data=True, edge_color=(0.8, 0.8, 0.8), edge_size=0.2,
                   text_size=0.042, text_color=(0, 0, 0),
                   remove_nodes=False, nbunch=[],
                   remove_edges=False, ebunch=[]):
        """
        Plots coloured balls at the region centres of connectivity, colour and
        size is determined by a vector of length number of regions (node_data).
        
        Optional: adds the connections between pair of nodes.
        
        """

        mlab.figure(1, bgcolor=bgcolor)

        # Get graph
        grf = nx.from_numpy_matrix(numpy.matrix(connectivity.weights))

        # Get the subgraph of nodes in nbunch
        if remove_nodes:
            grf.remove_nodes_from([n for n in grf if n not in set(nbunch)])
            # G.remove_nodes_from([node for node,degree in G.degree().items() if degree < 2])

        if remove_edges:
            grf.remove_edges_from([e for e in grf.edges() if e not in ebunch])

        # scalar colors
        if node_data is not None:
            scalars = node_data
            # mlab.colorbar(orientation="vertical")
        else:
            scalars = numpy.array(grf.nodes()) * 20

        pts = mlab.points3d(connectivity.centres[:, 0],
                            connectivity.centres[:, 1],
                            connectivity.centres[:, 2],
                            scalars,
                            # mask_points=1,
                            scale_factor=node_size,
                            scale_mode='none',
                            colormap=balls_colormap,
                            resolution=5,
                            opacity=0.01)

        if labels:
            if labels_indices is not None:
                for i, (idx) in enumerate(labels_indices):
                    x = connectivity.centres[idx, 0]
                    y = connectivity.centres[idx, 1]
                    z = connectivity.centres[idx, 2]
                    label = mlab.text(x, y, connectivity.region_labels[idx],
                                      z=z,
                                      width=text_size,
                                      name=str(connectivity.region_labels[idx]),
                                      color=text_color)
                    label.property.shadow = False
            else:
                for i, (x, y, z) in enumerate(connectivity.centres):
                    label = mlab.text(x, y, connectivity.region_labels[i],
                                      z=z,
                                      width=text_size,
                                      name=str(connectivity.region_labels[i]),
                                      color=text_color)
                    label.property.shadow = False

        if edge_data:
            pts.mlab_source.dataset.lines = numpy.array(grf.edges())
            tube = mlab.pipeline.tube(pts, tube_radius=edge_size)
            mlab.pipeline.surface(tube, color=edge_color, representation='wireframe', opacity=0.3)

        # mlab.show()

        # stop the scene
        # mlab.show(stop=True)


    def connectivity_3d(connectivity, edge_cutoff=None):
        """
        Plots a 3D representation of the delayed-connectivity structure.
        See Fig. 3 in (Knock et al 2009)

        [Nodes x Nodes x Delays]
        
        Original script can be found at: 
        BrainNetworkModels_3.1/PlottingTools/PlotConnectivity3D.m
        
        """

        mlab.figure(figure="Connectivity 3D", bgcolor=(0.0, 0.0, 0.0))

        if connectivity.delays is None:
            connectivity.configure()

        n_reg_half = connectivity.number_of_regions // 2

        min_d = connectivity.delays.min()
        max_d = connectivity.delays.max()
        step_d = (max_d - min_d) / 10.

        if edge_cutoff is None:
            edge_cutoff = connectivity.weights.min()

        # Loop over connectivity matrix, colouring and one cube per matrix element
        k = []
        d = []
        m = []
        s = []
        for k_idx in range(n_reg_half):
            for m_idx in range(n_reg_half):
                if connectivity.weights[k_idx, m_idx] != 0:
                    if k_idx != m_idx:

                        if connectivity.weights[k_idx, m_idx] > edge_cutoff:
                            k.append(k_idx + 2.)
                            d.append(connectivity.delays[k_idx, m_idx] + step_d)
                            m.append(m_idx + 2.0)
                            s.append(connectivity.weights[k_idx, m_idx])
        mlab.points3d(k, d, m, s, mode='cube')
        mlab.show(stop=True)


except ImportError:
    LOG.warning(
        "Mayavi is needed for this demo but due to sizing and packaging constraints we are not distributing it. "
        "If you want to see the actual plot you should use the github version and install all the required "
        "dependencies as described here: (advanced users only)"
        "http://docs.thevirtualbrain.com/manuals/ContributorsManual/ContributorsManual.html#the-unaided-setup")



"""
Mean field model based on Master equation about adaptative exponential leacky integrate and fire neurons population
"""
import numpy
import scipy.special as sp_spec
from tvb.basic.neotraits.api import NArray, Range, Final, List
from tvb.simulator.models.base import Model


class ZerlautFirstOrder(Model):
    r"""
    **References**:
    .. [ZD_2018]  Zerlaut, Y., Chemla, S., Chavane, F. et al. *Modeling mesoscopic cortical dynamics using a mean-field
    model of conductance-based networks of adaptive
    exponential integrate-and-fire neurons*,
    J Comput Neurosci (2018) 44: 45. https://doi-org.lama.univ-amu.fr/10.1007/s10827-017-0668-2
    .. [MV_2018]  Matteo di Volo, Alberto Romagnoni, Cristiano Capone, Alain Destexhe (2018)
    *Mean-field model for the dynamics of conductance-based networks of excitatory and inhibitory spiking neurons
    with adaptation*, bioRxiv, doi: https://doi.org/10.1101/352393

    Used Eqns 4 from [MV_2018]_ in ``dfun``.

    The default parameters are taken from table 1 of [ZD_2018]_, pag.47 and modify for the adaptation [MV_2018]
    +---------------------------+------------+
    |                 Table 1                |
    +--------------+------------+------------+
    |Parameter     |  Value     | Unit       |
    +==============+============+============+
    |             cellular property          |
    +--------------+------------+------------+
    | g_L          |   10.00    |   nS       |
    +--------------+------------+------------+
    | E_L_e        |  -67.00    |   mV       |
    +--------------+------------+------------+
    | E_L_i        |  -63.00    |   mV       |
    +--------------+------------+------------+
    | C_m          |   150.0    |   pF       |
    +--------------+------------+------------+
    | b            |   60.0     |   nS       |
    +--------------+------------+------------+
    | tau_w        |   500.0    |   ms       |
    +--------------+------------+------------+
    | T            |   5.0      |   ms       |
    +--------------+------------+------------+
    |          synaptic properties           |
    +--------------+------------+------------+
    | E_e          |    0.0     | mV         |
    +--------------+------------+------------+
    | E_i          |   -80.0    | mV         |
    +--------------+------------+------------+
    | Q_e          |    1.0     | nS         |
    +--------------+------------+------------+
    | Q_i          |    5.0     | nS         |
    +--------------+------------+------------+
    | tau_e        |    5.0     | ms         |
    +--------------+------------+------------+
    | tau_i        |    5.0     | ms         |
    +--------------+------------+------------+
    |          numerical network             |
    +--------------+------------+------------+
    | N_tot        |  10000     |            |
    +--------------+------------+------------+
    | p_connect    |    5.0 %   |            |
    +--------------+------------+------------+
    | g            |   20.0 %   |            |
    +--------------+------------+------------+
    |external_input|    0.001   | Hz         |
    +--------------+------------+------------+

    The default coefficients of the transfert function are taken from table I of [MV_2018]_, pag.49
    +-------------+-------------+-------------+-------------+-------------+-------------+-------------+-------------+-------------+-------------+
    |      excitatory cell      |
    +-------------+-------------+-------------+-------------+-------------+-------------+-------------+-------------+-------------+-------------+
    |  -4.98e-02  |   5.06e-03  |  -2.5e-02   |   1.4e-03   |  -4.1e-04   |   1.05e-02  |  -3.6e-02   |   7.4e-03   |   1.2e-03   |  -4.07e-02  |
    +-------------+-------------+-------------+-------------+-------------+-------------+-------------+-------------+-------------+-------------+
    |      inhibitory cell      |
    +-------------+-------------+-------------+-------------+-------------+-------------+-------------+-------------+-------------+-------------+
    |  -5.14e-02  |   4.0e-03   |  -8.3e-03   |   2.0e-04   |  -5.0e-04   |   1.4e-03   |  -1.46e-02  |   4.5e-03   |   2.8e-03   |  -1.53e-02  |
    +-------------+-------------+-------------+-------------+-------------+-------------+-------------+-------------+-------------+-------------+

    The models (:math:`E`, :math:`I`) phase-plane, including a representation of
    the vector field as well as its nullclines, using default parameters, can be
    seen below.

    The general formulation for the \textit{\textbf{Zerlaut_adaptation_first_order}} model as a
    dynamical unit at a node $k$ in a BNM with $l$ nodes reads:

    .. math::
            T\dot{E}_k &= F_e-E_k  \\
            T\dot{I}_k &= F_i-I_k  \\
            dot{W}_k &= W_k/tau_w-b*E_k  \\
            F_\lambda = Erfc(V^{eff}_{thre}-\mu_V/\sqrt(2)\sigma_V)

    """

    # Define traited attributes for this model, these represent possible kwargs.
    g_L = NArray(
        label=":math:`g_{L}`",
        default=numpy.array([10.]),  # 10 nS by default, i.e. ~ 100MOhm input resitance at rest
        domain=Range(lo=0.1, hi=100.0, step=0.1),  # 0.1nS would be a very small cell, 100nS a very big one
        doc="""leak conductance [nS]""")

    E_L_e = NArray(
        label=":math:`E_{L}`",
        default=numpy.array([-67.0]),
        domain=Range(lo=-90.0, hi=-60.0, step=0.1),  # resting potential, usually between -85mV and -65mV
        doc="""leak reversal potential for excitatory [mV]""")

    E_L_i = NArray(
        label=":math:`E_{L}`",
        default=numpy.array([-63.0]),
        domain=Range(lo=-90.0, hi=-60.0, step=0.1),  # resting potential, usually between -85mV and -65mV
        doc="""leak reversal potential for inhibitory [mV]""")

    # N.B. Not independent of g_L, C_m should scale linearly with g_L
    C_m = NArray(
        label=":math:`C_{m}`",
        default=numpy.array([150]),
        domain=Range(lo=10.0, hi=500.0, step=10.0),  # 20pF very small cell, 400pF very
        doc="""membrane capacitance [pF]""")

    b = NArray(
        label=":math:`b`",
        default=numpy.array([60.0]),
        domain=Range(lo=0.0, hi=150.0, step=1.0),
        doc="""Adaptation [nS]""")

    tau_w = NArray(
        label=":math:`tau_w`",
        default=numpy.array([500.0]),
        domain=Range(lo=5.0, hi=1000.0, step=1.0),
        doc="""Adaptation time constant [ms]""")

    E_e = NArray(
        label=r":math:`E_e`",
        default=numpy.array([0.0]),
        domain=Range(lo=-20., hi=20., step=0.01),
        doc="""excitatory reversal potential [mV]""")

    E_i = NArray(
        label=":math:`E_i`",
        default=numpy.array([-80.0]),
        domain=Range(lo=-100.0, hi=-60.0, step=1.0),
        doc="""inhibitory reversal potential [mV]""")

    Q_e = NArray(
        label=r":math:`Q_e`",
        default=numpy.array([1.0]),
        domain=Range(lo=0.0, hi=5.0, step=0.1),
        doc="""excitatory quantal conductance [nS]""")

    Q_i = NArray(
        label=r":math:`Q_i`",
        default=numpy.array([5.0]),
        domain=Range(lo=0.0, hi=10.0, step=0.1),
        doc="""inhibitory quantal conductance [nS]""")

    tau_e = NArray(
        label=":math:`\tau_e`",
        default=numpy.array([5.0]),
        domain=Range(lo=1.0, hi=10.0, step=1.0),
        doc="""excitatory decay [ms]""")

    tau_i = NArray(
        label=":math:`\tau_i`",
        default=numpy.array([5.0]),
        domain=Range(lo=0.5, hi=10.0, step=0.01),
        doc="""inhibitory decay [ms]""")

    N_tot = NArray(
        dtype=numpy.int,
        label=":math:`N_{tot}`",
        default=numpy.array([10000]),
        domain=Range(lo=1000, hi=50000, step=1000),
        doc="""cell number""")

    p_connect = NArray(
        label=":math:`\epsilon`",
        default=numpy.array([0.05]),
        domain=Range(lo=0.001, hi=0.2, step=0.001),  # valid only for relatively sparse connectivities
        doc="""connectivity probability""")

    g = NArray(
        label=":math:`g`",
        default=numpy.array([0.2]),
        domain=Range(lo=0.01, hi=0.4, step=0.01),  # inhibitory cell number never overcomes excitatory ones
        doc="""fraction of inhibitory cells""")

    T = NArray(
        label=":math:`T`",
        default=numpy.array([5.0]),
        domain=Range(lo=1., hi=20.0, step=0.1),
        doc="""time scale of describing network activity""")

    P_e = NArray(
        label=":math:`P_e`",  # TODO need to check the size of the array when it's used
        default=numpy.array([-4.98e-02, 5.06e-03, -2.5e-02, 1.4e-03,
                             -4.1e-04, 1.05e-02, -3.6e-02, 7.4e-03,
                             1.2e-03, -4.07e-02]),
        doc="""Polynome of excitatory phenomenological threshold (order 9)""")

    P_i = NArray(
        label=":math:`P_i`",  # TODO need to check the size of the array when it's used
        default=numpy.array([-5.14e-02, 4.0e-03, -8.3e-03, 2.0e-04,
                             -5.0e-04, 1.4e-03, -1.46e-02, 4.5e-03,
                             2.8e-03, -1.53e-02]),
        doc="""Polynome of inhibitory phenomenological threshold (order 9)""")

    external_input = NArray(
        label=":math:`\nu_e^{drive}`",
        default=numpy.array([0.0001]),
        domain=Range(lo=0.00, hi=0.1, step=0.001),
        doc="""external drive""")

    # Used for phase-plane axis ranges and to bound random initial() conditions.
    state_variable_range = Final(
        {
            "E": numpy.array([0.0, 0.1]),  # actually the 100Hz should be replaced by 1/T_refrac
            "I": numpy.array([0.0, 0.1]),
            "W": numpy.array([0.0, 100.0])
        },
        label="State Variable ranges [lo, hi]",
        doc="""The values for each state-variable should be set to encompass
        the expected dynamic range of that state-variable for the current
        parameters, it is used as a mechanism for bounding random initial
        conditions when the simulation isn't started from an explicit history,
        it is also provides the default range of phase-plane plots.\n
        E: firing rate of excitatory population in KHz\n
        I: firing rate of inhibitory population in KHz\n
        W: level of adaptation
        """)

    variables_of_interest = List(
        of=str,
        label="Variables watched by Monitors",
        choices=("E", "I", "W"),
        default=("E",),
        doc="""This represents the default state-variables of this Model to be
               monitored. It can be overridden for each Monitor if desired. The
               corresponding state-variable indices for this model are :math:`E = 0`,
               :math:`I = 1` and :math:`W = 2`.""")

    state_variables = 'E I W'.split()
    _nvar = 3
    cvar = numpy.array([0, 1, 2], dtype=numpy.int32)

    def dfun(self, state_variables, coupling, local_coupling=0.00):
        r"""
        .. math::
            T \dot{\nu_\mu} &= -F_\mu(\nu_e,\nu_i) + \nu_\mu ,\all\mu\in\{e,i\}\\
            dot{W}_k &= W_k/tau_w-b*E_k  \\

        """
        E = state_variables[0, :]
        I = state_variables[1, :]
        W = state_variables[2, :]
        derivative = numpy.empty_like(state_variables)

        # long-range coupling
        c_0 = coupling[0, :]

        # short-range (local) coupling
        lc_E = local_coupling * E
        lc_I = local_coupling * I

        # Excitatory firing rate derivation
        derivative[0] = (self.TF_excitatory(E+c_0+lc_E+self.external_input, I+lc_I+self.external_input,W)-E)/self.T
        # Inhibitory firing rate derivation
        derivative[1] = (self.TF_inhibitory(E+lc_E+self.external_input, I+lc_I+self.external_input,W)-I)/self.T
        # Adaptation
        derivative[2] = -W/self.tau_w+self.b*E

        return derivative

    def TF_excitatory(self, fe, fi, W):
        """
        transfer function for excitatory population
        :param fe: firing rate of excitatory population
        :param fi: firing rate of inhibitory population
        :param W: level of adaptation
        :return: result of transfer function
        """
        return self.TF(fe, fi, W, self.P_e, self.E_L_e)

    def TF_inhibitory(self, fe, fi, W):
        """
        transfer function for inhibitory population
        :param fe: firing rate of excitatory population
        :param fi: firing rate of inhibitory population
        :param W: level of adaptation
        :return: result of transfer function
        """
        return self.TF(fe, fi, W, self.P_i, self.E_L_i)

    def TF(self, fe, fi, W, P, E_L):
        """
        transfer function for inhibitory population
        Inspired from the next repository :
        https://github.com/yzerlaut/notebook_papers/tree/master/modeling_mesoscopic_dynamics
        :param fe: firing rate of excitatory population
        :param fi: firing rate of inhibitory population
        :param W: level of adaptation
        :param P: Polynome of neurons phenomenological threshold (order 9)
        :param E_L: leak reversal potential
        :return: result of transfer function
        """
        mu_V, sigma_V, T_V = self.get_fluct_regime_vars(fe, fi, W, self.Q_e, self.tau_e, self.E_e,
                                                           self.Q_i, self.tau_i, self.E_i,
                                                           self.g_L, self.C_m, E_L, self.N_tot,
                                                           self.p_connect, self.g)
        V_thre = self.threshold_func(mu_V, sigma_V, T_V*self.g_L/self.C_m,
                                     P[0], P[1], P[2], P[3], P[4], P[5], P[6], P[7], P[8], P[9])
        V_thre *= 1e3  # the threshold need to be in mv and not in Volt
        f_out = self.estimate_firing_rate(mu_V, sigma_V, T_V, V_thre)
        return f_out

    @staticmethod
    def get_fluct_regime_vars(Fe, Fi, W, Q_e, tau_e, E_e, Q_i, tau_i, E_i, g_L, C_m, E_L, N_tot, p_connect, g):
        """
        Compute the mean characteristic of neurons.
        Inspired from the next repository :
        https://github.com/yzerlaut/notebook_papers/tree/master/modeling_mesoscopic_dynamics
        :param Fe: firing rate of excitatory population
        :param Fi: firing rate of inhibitory population
        :param W: level of adaptation
        :param Q_e: excitatory quantal conductance
        :param tau_e: excitatory decay
        :param E_e: excitatory reversal potential
        :param Q_i: inhibitory quantal conductance
        :param tau_i: inhibitory decay
        :param E_i: inhibitory reversal potential
        :param E_L: leakage reversal voltage of neurons
        :param g_L: leak conductance
        :param C_m: membrane capacitance
        :param E_L: leak reversal potential
        :param N_tot: cell number
        :param p_connect: connectivity probability
        :param g: fraction of inhibitory cells
        :return: mean and variance of membrane voltage of neurons and autocorrelation time constant
        """
        # firing rate
        # 1e-6 represent spontaneous release of synaptic neurotransmitter or some intrinsic currents of neurons
        fe = (Fe+1e-6)*(1.-g)*p_connect*N_tot
        fi = (Fi+1e-6)*g*p_connect*N_tot

        # conductance fluctuation and effective membrane time constant
        mu_Ge, mu_Gi = Q_e*tau_e*fe, Q_i*tau_i*fi  # Eqns 5 from [MV_2018]
        mu_G = g_L+mu_Ge+mu_Gi  # Eqns 6 from [MV_2018]
        T_m = C_m/mu_G # Eqns 6 from [MV_2018]

        # membrane potential
        mu_V = (mu_Ge*E_e+mu_Gi*E_i+g_L*E_L-W)/mu_G  # Eqns 7 from [MV_2018]
        # post-synaptic membrane potential event s around muV
        U_e, U_i = Q_e/mu_G*(E_e-mu_V), Q_i/mu_G*(E_i-mu_V)
        # Standard deviation of the fluctuations
        # Eqns 8 from [MV_2018]
        sigma_V = numpy.sqrt(fe*(U_e*tau_e)**2/(2.*(tau_e+T_m))+fi*(U_i*tau_i)**2/(2.*(tau_i+T_m)))
        # Autocorrelation-time of the fluctuations Eqns 9 from [MV_2018]
        T_V_numerator = (fe*(U_e*tau_e)**2 + fi*(U_i*tau_i)**2)
        T_V_denominator = (fe*(U_e*tau_e)**2/(tau_e+T_m) + fi*(U_i*tau_i)**2/(tau_i+T_m))
        T_V = numpy.divide(T_V_numerator, T_V_denominator, out=numpy.ones_like(T_V_numerator),
                           where=T_V_denominator != 0.0)
        return mu_V, sigma_V, T_V

    @staticmethod
    def threshold_func(muV, sigmaV, TvN, P0, P1, P2, P3, P4, P5, P6, P7, P8, P9):
        """
        The threshold function of the neurons
        :param muV: mean of membrane voltage
        :param sigmaV: variance of membrane voltage
        :param TvN: autocorrelation time constant
        :param P: Fitted coefficients of the transfer functions
        :return: threshold of neurons
        """
        # Normalization factors page 48 after the equation 4 from [ZD_2018]
        muV0, DmuV0 = -60.0, 10.0
        sV0, DsV0 = 4.0, 6.0
        TvN0, DTvN0 = 0.5, 1.
        V = (muV-muV0)/DmuV0
        S = (sigmaV-sV0)/DsV0
        T = (TvN-TvN0)/DTvN0
        # Eqns 11 from [MV_2018]
        return P0 + P1*V + P2*S + P3*T + P4*V**2 + P5*S**2 + P6*T**2 + P7*V*S + P8*V*T + P9*S*T

    @staticmethod
    def estimate_firing_rate(muV, sigmaV, Tv, Vthre):
        # Eqns 10 from [MV_2018]
        return sp_spec.erfc((Vthre-muV) / (numpy.sqrt(2)*sigmaV)) / (2*Tv)


class ZerlautSecondOrder(ZerlautFirstOrder):
    r"""
    **References**:
    .. [ZD_2018]  Zerlaut, Y., Chemla, S., Chavane, F. et al. *Modeling mesoscopic cortical dynamics using a mean-field
    model of conductance-based networks of adaptive
    exponential integrate-and-fire neurons*,
    J Comput Neurosci (2018) 44: 45. https://doi-org.lama.univ-amu.fr/10.1007/s10827-017-0668-2
    .. [MV_2018]  Matteo di Volo, Alberto Romagnoni, Cristiano Capone, Alain Destexhe (2018)
    *Mean-field model for the dynamics of conductance-based networks of excitatory and inhibitory spiking neurons
    with adaptation*, bioRxiv, doi: https://doi.org/10.1101/352393

    Used Eqns 4 from [MV_2018]_ in ``dfun``.

    (See Zerlaut_adaptation_first_order for the default value)

    The models (:math:`E`, :math:`I`) phase-plane, including a representation of
    the vector field as well as its nullclines, using default parameters, can be
    seen below.

    The general formulation for the \textit{\textbf{Zerlaut_adaptation_second_order}} model as a
    dynamical unit at a node $k$ in a BNM with $l$ nodes reads:

    .. math::
        \forall \mu,\lambda,\eta \in \{e,i\}^3\, ,
        \left\{
        \begin{split}
        T \, \frac{\partial \nu_\mu}{\partial t} = & (\mathcal{F}_\mu - \nu_\mu )
        + \frac{1}{2} \, c_{\lambda \eta} \,
        \frac{\partial^2 \mathcal{F}_\mu}{\partial \nu_\lambda \partial \nu_\eta} \\
        T \, \frac{\partial c_{\lambda \eta} }{\partial t}  =  & A_{\lambda \eta} +
        (\mathcal{F}_\lambda - \nu_\lambda ) \, (\mathcal{F}_\eta - \nu_\eta ) + \\
        & c_{\lambda \mu} \frac{\partial \mathcal{F}_\mu}{\partial \nu_\lambda} +
        c_{\mu \eta} \frac{\partial \mathcal{F}_\mu}{\partial \nu_\eta}
        - 2  c_{\lambda \eta}
        \end{split}
        \right.
        dot{W}_k &= W_k/tau_w-b*E_k  \\

        with:
        A_{\lambda \eta} =
        \left\{
        \begin{split}
        \frac{\mathcal{F}_\lambda \, (1/T - \mathcal{F}_\lambda)}{N_\lambda}
        \qquad & \textrm{if  } \lambda=\eta \\
        0 \qquad & \textrm{otherwise}
        \end{split}
        \right.
    """

    #  Used for phase-plane axis ranges and to bound random initial() conditions.
    state_variable_range = Final(
        {
            "E": numpy.array([0.0, 0.1]),  # actually the 100Hz should be replaced by 1/T_refrac
            "I": numpy.array([0.0, 0.1]),
            "C_ee": numpy.array([0.0, 0.0]),  # variance is positive or null
            "C_ei": numpy.array([0.0, 0.0]),  # the co-variance is in [-c_ee*c_ii,c_ee*c_ii]
            "C_ii": numpy.array([0.0, 0.0]),  # variance is positive or null
            "W": numpy.array([0.0, 100.0])
        },
        label="State Variable ranges [lo, hi]",
        doc="""The values for each state-variable should be set to encompass
        the expected dynamic range of that state-variable for the current
        parameters, it is used as a mechanism for bounding random inital
        conditions when the simulation isn't started from an explicit history,
        it is also provides the default range of phase-plane plots.\n
        E: firing rate of excitatory population in KHz\n
        I: firing rate of inhibitory population in KHz\n
        C_ee: the variance of the excitatory population activity \n
        C_ei: the covariance between the excitatory and inhibitory population activities (always symetric) \n
        C_ie: the variance of the inhibitory population activity \n
        W: level of adaptation
        """)

    variables_of_interest = List(
        of=str,
        label="Variables watched by Monitors",
        choices=("E", "I", "C_ee", "C_ei", "C_ii"),
        default=("E",),
        doc="""This represents the default state-variables of this Model to be
               monitored. It can be overridden for each Monitor if desired. The
               corresponding state-variable indices for this model are :math:`E = 0`,
               :math:`I = 1`, :math:`C_ee = 2`, :math:`C_ei = 3`, :math:`C_ii = 4` and :math:`W = 5`.""")

    state_variables = 'E I C_ee C_ei C_ii W'.split()
    _nvar = 6
    cvar = numpy.array([0, 1, 2, 3, 4, 5], dtype=numpy.int32)

    def dfun(self, state_variables, coupling, local_coupling=0.00):
        r"""
        .. math::
            \forall \mu,\lambda,\eta \in \{e,i\}^3\, ,
            \left\{
            \begin{split}
            T \, \frac{\partial \nu_\mu}{\partial t} = & (\mathcal{F}_\mu - \nu_\mu )
            + \frac{1}{2} \, c_{\lambda \eta} \,
            \frac{\partial^2 \mathcal{F}_\mu}{\partial \nu_\lambda \partial \nu_\eta} \\
            T \, \frac{\partial c_{\lambda \eta} }{\partial t}  =  & A_{\lambda \eta} +
            (\mathcal{F}_\lambda - \nu_\lambda ) \, (\mathcal{F}_\eta - \nu_\eta ) + \\
            & c_{\lambda \mu} \frac{\partial \mathcal{F}_\mu}{\partial \nu_\lambda} +
            c_{\mu \eta} \frac{\partial \mathcal{F}_\mu}{\partial \nu_\eta}
            - 2  c_{\lambda \eta}
            \end{split}
            \right.
            dot{W}_k &= W_k/tau_w-b*E_k  \\

            with:
            A_{\lambda \eta} =
            \left\{
            \begin{split}
            \frac{\mathcal{F}_\lambda \, (1/T - \mathcal{F}_\lambda)}{N_\lambda}
            \qquad & \textrm{if  } \lambda=\eta \\
            0 \qquad & \textrm{otherwise}
            \end{split}
            \right.

        """
        N_e = self.N_tot * (1-self.g)
        N_i = self.N_tot * self.g

        E = state_variables[0, :]
        I = state_variables[1, :]
        C_ee = state_variables[2, :]
        C_ei = state_variables[3, :]
        C_ii = state_variables[4, :]
        W = state_variables[5,:]
        derivative = numpy.empty_like(state_variables)

        # long-range coupling
        c_0 = coupling[0, :]

        # short-range (local) coupling
        lc_E = local_coupling * E
        lc_I = local_coupling * I

        E_input_excitatory = E+c_0+lc_E+self.external_input
        E_input_inhibitory = E+lc_E+self.external_input
        I_input_excitatory = I+lc_I+self.external_input
        I_input_inhibitory = I+lc_I+self.external_input

        # Transfer function of excitatory and inhibitory neurons
        _TF_e = self.TF_excitatory(E_input_excitatory, I_input_excitatory, W)
        _TF_i = self.TF_inhibitory(E_input_inhibitory, I_input_inhibitory, W)

        # Derivatives taken numerically : use a central difference formula with spacing `dx`
        def _diff_fe(TF, fe, fi, W, df=1e-7):
            return (TF(fe+df, fi, W)-TF(fe-df, fi, W))/(2*df*1e3)

        def _diff_fi(TF, fe, fi, W, df=1e-7):
            return (TF(fe, fi+df, W)-TF(fe, fi-df, W))/(2*df*1e3)

        def _diff2_fe_fe_e(fe, fi, W, df=1e-7):
            TF = self.TF_excitatory
            return (TF(fe+df, fi, W)-2*_TF_e+TF(fe-df, fi, W))/((df*1e3)**2)

        def _diff2_fe_fe_i(fe, fi, W, df=1e-7):
            TF = self.TF_inhibitory
            return (TF(fe+df, fi, W)-2*_TF_i+TF(fe-df, fi, W))/((df*1e3)**2)

        def _diff2_fi_fe(TF, fe, fi, W, df=1e-7):
            return (_diff_fi(TF, fe+df, fi, W)-_diff_fi(TF, fe-df, fi, W))/(2*df*1e3)

        def _diff2_fe_fi(TF, fe, fi, W, df=1e-7):
            return (_diff_fe(TF, fe, fi+df, W)-_diff_fe(TF, fe, fi-df, W))/(2*df*1e3)

        def _diff2_fi_fi_e(fe, fi, W, df=1e-7):
            TF = self.TF_excitatory
            return (TF(fe, fi+df, W)-2*_TF_e+TF(fe, fi-df, W))/((df*1e3)**2)

        def _diff2_fi_fi_i(fe, fi, W, df=1e-7):
            TF = self.TF_inhibitory
            return (TF(fe, fi+df, W)-2*_TF_i+TF(fe, fi-df, W))/((df*1e3)**2)

        #Precompute some result
        _diff_fe_TF_e = _diff_fe(self.TF_excitatory, E_input_excitatory, I_input_excitatory, W)
        _diff_fe_TF_i = _diff_fe(self.TF_inhibitory, E_input_inhibitory, I_input_inhibitory, W)
        _diff_fi_TF_e = _diff_fi(self.TF_excitatory, E_input_excitatory, I_input_excitatory, W)
        _diff_fi_TF_i = _diff_fi(self.TF_inhibitory, E_input_inhibitory, I_input_inhibitory, W)

        # equation is inspired from github of Zerlaut :
        # https://github.com/yzerlaut/notebook_papers/blob/master/modeling_mesoscopic_dynamics/mean_field/master_equation.py
        # Excitatory firing rate derivation
        derivative[0] = (_TF_e - E
                         + .5*C_ee*_diff2_fe_fe_e(E_input_excitatory, I_input_excitatory, W)
                         + .5*C_ei*_diff2_fe_fi(self.TF_excitatory, E_input_excitatory, I_input_excitatory, W)
                         + .5*C_ei*_diff2_fi_fe(self.TF_excitatory, E_input_excitatory, I_input_excitatory, W)
                         + .5*C_ii*_diff2_fi_fi_e(E_input_excitatory, I_input_excitatory, W)
                         )/self.T
        # Inhibitory firing rate derivation
        derivative[1] = (_TF_i - I
                         + .5*C_ee*_diff2_fe_fe_i(E_input_inhibitory, I_input_inhibitory, W)
                         + .5*C_ei*_diff2_fe_fi(self.TF_inhibitory, E_input_inhibitory, I_input_inhibitory, W)
                         + .5*C_ei*_diff2_fi_fe(self.TF_inhibitory, E_input_inhibitory, I_input_inhibitory, W)
                         + .5*C_ii*_diff2_fi_fi_i(E_input_inhibitory, I_input_inhibitory, W)
                         )/self.T
        # Covariance excitatory-excitatory derivation
        derivative[2] = (_TF_e*(1./self.T-_TF_e)/N_e
                         + (_TF_e-E)**2
                         + 2.*C_ee*_diff_fe_TF_e
                         + 2.*C_ei*_diff_fi_TF_i
                         - 2.*C_ee
                         )/self.T
        # Covariance excitatory-inhibitory or inhibitory-excitatory derivation
        derivative[3] = ((_TF_e-E)*(_TF_i-I)
                         + C_ee*_diff_fe_TF_e
                         + C_ei*_diff_fe_TF_i
                         + C_ei*_diff_fi_TF_e
                         + C_ii*_diff_fi_TF_i
                         - 2.*C_ei
                         )/self.T
        # Covariance inhibitory-inhibitory derivation
        derivative[4] = (_TF_i*(1./self.T-_TF_i)/N_i
                         + (_TF_i-I)**2
                         + 2.*C_ii*_diff_fi_TF_i
                         + 2.*C_ei*_diff_fe_TF_e
                         - 2.*C_ii
                         )/self.T
        # Adaptation
        derivative[5] = -W/self.tau_w+self.b*E

        return derivative




# -*- coding: utf-8 -*-
#
#
#  TheVirtualBrain-Scientific Package. This package holds all simulators, and
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2017, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

import pyopencl
import pyopencl.array
import numpy
import os
from .models import CLComponent,CLModel
from ..models.oscillator import Generic2dOscillator,Kuramoto
from ..models.jansen_rit import JansenRit, ZetterbergJansen
from ..models.linear import Linear
from ..models.larter_breakspear import LarterBreakspear
from ..models.epileptor import Epileptor
from ..models.hopfield import Hopfield
from ..models.stefanescu_jirsa import ReducedSetFitzHughNagumo, ReducedSetHindmarshRose
from ..models.wilson_cowan import WilsonCowan


class CL_Kuramoto(Kuramoto, CLModel):
    n_states = 1
    _opencl_ordered_params = "omega".split()
    _opencl_program_source_file = "Kuramoto.cl"
    _opencl_program_source = """
    __kernel void dfun(__global float *state, __global float *coupling,
                   __global float *param, __global float *deriv)
    {
        int i = get_global_id(0), n = get_global_size(0);

        // this is boilerplate and could be generated
        float S=state[i], omega=param[i];
        float I = coupling[i];
        deriv[i] = omega + I;
    }
    """
    # with open(_opencl_program_source_file, "r") as fd:
    #     _opencl_program_source = fd.read()


class CL_Generic2D(Generic2dOscillator, CLModel):

    n_states = 2
    _opencl_ordered_params = "tau a b c I d e f g alpha beta gamma".split()
    _opencl_program_source_file = "Generic2D.cl"
    _opencl_program_source = """
    //Generic2dOscillator
    #define indexNum(cur,totalN) cur*n+i
    __kernel void dfun(__global float *state, __global float *coupling,
                       __global float *param, __global float *deriv)
    {
        int i = get_global_id(0), n = get_global_size(0);

        // this is boilerplate and could be generated
        float V = state[i*2], W = state[i*2+1];
        //tau, a, b, c, I, d, e, f, g, alpha, beta, gamma
        float c_0 = coupling[i];
        float tau = param[n+i], a = param[2*n+i],
        b = param[3*n+i], c = param[4*n+i], I= param[5*n+i], d = param[6*n+i],
        e = param[7*n+i], f = param[8*n+i], g= param[9*n+i], alpha= param[10*n+i],
        beta = param[11*n+i], gamma = param[12*n+i];

        deriv[indexNum(0,2)] = d * tau * (alpha * W - f * V*V*V + e * V*V + g * V + gamma * I + gamma *c_0);
        deriv[indexNum(1,2)] = d * (a + b * V + c * V*V - beta * W) / tau;
    }"""
    # with open(os.path.join('/',_opencl_program_source_file), "r") as fd:
    #       _opencl_program_source = fd.read()

class CL_Jansen_rit(JansenRit, CLModel):
    n_states = 6
    _opencl_ordered_params = "A B a b v0 nu_max r J a_1 a_2 a_3 a_4 p_min p_max mu".split()
    _opencl_program_source_file = "Jansen_rit.cl"
    _opencl_program_source = """
    #define indexNum(cur,totalN) cur*n+i // consider i*totalN+cur
    __kernel void dfun(__global float *state, __global float *coupling,
                       __global float *param, __global float *deriv)
    {
        int i = get_global_id(0), n = get_global_size(0);

        // this is boilerplate and could be generated
        float y0 = state[i*6], y1 = state[i*6+1],y2 = state[i*6+2],
        y3 = state[i*6+3], y4 = state[i*6+4], y5 = state[i*6+5];
        float c = coupling[i]; ;
        // nu_max  r  v0  a  a_1  a_2  a_3  a_4  A  b  B  J  mu //total 13 parameters
        float nu_max = param[indexNum(0,13)];
        float  r = param[indexNum(1,13)];
        float  v0 = param[indexNum(2,13)];
        float  a = param[indexNum(3,13)];
        float  a_1 = param[indexNum(4,13)];
        float  a_2 = param[indexNum(5,13)];
        float  a_3 = param[indexNum(6,13)];
        float  a_4 = param[indexNum(7,13)];
        float  A = param[indexNum(8,13)];
        float  b = param[indexNum(9,13)];
        float  B = param[indexNum(10,13)];
        float  J = param[indexNum(11,13)];
        float  mu = param[indexNum(12,13)];

        float src = y1 - y2;

        float sigm_y1_y2 = 2.0 * nu_max / (1.0 + exp(r * (v0 - (y1 - y2))));
        float sigm_y0_1 = 2.0 * nu_max / (1.0 + exp(r * (v0 - (a_1 * J * y0))));
        float sigm_y0_3 = 2.0 * nu_max / (1.0 + exp(r * (v0 - (a_3 * J * y0))));
        deriv[indexNum(0,6)] = y3;
        deriv[indexNum(1,6)] = y4;
        deriv[indexNum(2,6)] = y5;
        deriv[indexNum(3,6)] = A * a * sigm_y1_y2 - 2.0 * a * y3 - a * a * y0;
        deriv[indexNum(4,6)] = A * a * (mu + a_2 * J * sigm_y0_1 + c + src) - 2.0 * a * y4 - a * a * y1;
        deriv[indexNum(5,6)] = B * b * (a_4 * J * sigm_y0_3) - 2.0 * b * y5 - b *b * y2;
    }

    """


class CL_Zetterberg_Jasen(ZetterbergJansen, CLModel):
    n_states = 12
    _opencl_ordered_params = " He Hi ke ki e0 rho_2 rho_1 gamma_1 gamma_2 gamma_3 gamma_4 gamma_5 P U Q Heke Hiki ke_2 ki_2 keke kiki".split()
    _opencl_program_source_file = "Zetterberg_Jasen.cl"
    _opencl_program_source = """
    #define indexNum(cur,totalN) cur*n+i
    #define sigma_fun(sv) ( rho_1 *(rho_2 - sv) > 709 ? 0 : 2*e0 / (1+rho_1 *(rho_2 - sv)))


    __kernel void dfun(__global float *state, __global float *coupling,
                       __global float *param, __global float *deriv)
    {
        int i = get_global_id(0), n = get_global_size(0);
        int n_param = 12;
        // this is boilerplate and could be generated
        float v1 = state[indexNum(0,12)], y1 = state[indexNum(1,12)],v2 = state[indexNum(2,12)],
        y2 = state[indexNum(3,12)],v3 = state[indexNum(4,12)], y3 = state[indexNum(5,12)],
        v4 = state[indexNum(6,12)], y4 = state[indexNum(7,12)], v5 = state[indexNum(8,12)],
        y5 = state[indexNum(9,12)], v6 = state[indexNum(10,12)], v7 = state[indexNum(11,12)],
        c = coupling[i];
        //He Hi ke ki e0 rho_2 rho_1 gamma_1 gamma_2 gamma_3 gamma_4 gamma_5 P U Q Heke Hiki ke_2 ki_2 keke kiki gamma_1T gamma_2T gamma_3T
        float He = param[indexNum(0,24)];
        float Hi = param[indexNum(1,24)];
        float ke = param[indexNum(2,24)];
        float ki = param[indexNum(3,24)];
        float e0 = param[indexNum(4,24)];
        float rho_2 = param[indexNum(5,24)];
        float rho_1 = param[indexNum(6,24)];
        float gamma_1 = param[indexNum(7,24)];
        float gamma_2 = param[indexNum(8,24)];
        float gamma_3 = param[indexNum(9,24)];
        float gamma_4 = param[indexNum(10,24)];
        float gamma_5 = param[indexNum(11,24)];
        float P = param[indexNum(12,24)];
        float U = param[indexNum(13,24)];
        float Q = param[indexNum(14,24)];
        float Heke = param[indexNum(15,24)];
        float Hiki = param[indexNum(16,24)];
        float ke_2 = param[indexNum(17,24)];
        float ki_2 = param[indexNum(18,24)];
        float keke = param[indexNum(19,24)];
        float kiki = param[indexNum(20,24)];
        float gamma_1T = param[indexNum(21,24)];
        float gamma_2T = param[indexNum(22,24)];
        float gamma_3T = param[indexNum(23,24)];
        // TODO: implement local coupling
        float locol_coupling = 1;

        float coupled_input = c + 6*locol_coupling;

        deriv[indexNum(0,12)] = y1;
        deriv[indexNum(1,12)] = Heke * (gamma_1 * sigma_fun(v2 - v3) + gamma_1T * (U + coupled_input )) - ke_2 * y1 - keke * v1;
        // exc input to the pyramidal cells
        deriv[indexNum(2,12)] = y2;
        deriv[indexNum(3,12)] = Heke * (gamma_2 * sigma_fun(v1)      + gamma_2T * (P + coupled_input )) - ke_2 * y2 - keke * v2;
        // inh input to the pyramidal cells
        deriv[indexNum(4,12)] = y3;
        deriv[indexNum(5,12)] = Hiki * (gamma_4 * sigma_fun(v4 - v5)) - ki_2 * y3 - kiki * v3;
        deriv[indexNum(6,12)] = y4;
        // exc input to the inhibitory interneurons
        deriv[indexNum(7,12)] = Heke * (gamma_3 * sigma_fun(v2 - v3) + gamma_3T * (Q + coupled_input)) - ke_2 * y4 - keke * v4;
        deriv[indexNum(8,12)] = y5;
        // inh input to the inhibitory interneurons
        deriv[indexNum(9,12)] = Hiki * (gamma_5 * sigma_fun(v4 - v5)) - ki_2 * y5 - keke * v5;
        // aux variables (the sum gathering the postsynaptic inh & exc potentials)
        // pyramidal cells
        deriv[indexNum(10,12)] = y2 - y3;
        // inhibitory cells
        deriv[indexNum(11,12)] = y4 - y5;

    }

    """


class CL_Linear( Linear, CLModel ):
    n_states = 1
    _opencl_ordered_params = "gamma".split()
    _opencl_program_source_file = "Linear.cl"
    _opencl_program_source = """
    __kernel void dfun(__global float *state, __global float *coupling,
                       __global float *param, __global float *deriv)
    {
        int i = get_global_id(0), n = get_global_size(0);

        // this is boilerplate and could be generated
        float x = state[i];
        //gamma
        float c = coupling[i];
        float gamma = param[i];
        deriv[i] = gamma*x + x;
    }
    """

class CL_Hopfield( Hopfield, CLModel ):
    n_states = 2
    _opencl_ordered_params = "taux tauT dynamic".split()
    _opencl_program_source_file = "Hopfield.cl"
    _opencl_program_source = """
    #define indexNum(cur,totalN) cur*n+i
    __kernel void dfun(__global float *state, __global float *coupling,
                   __global float *param, __global float *deriv)
    {
        int i = get_global_id(0), n = get_global_size(0);

        // this is boilerplate and could be generated
        float x = state[indexNum(0,2)], theta=param[indexNum(1,2)];
        float c = coupling[i];
        // taux tauT dynamic
        float taux = param[indexNum(0,3)],tauT = param[indexNum(1,3)], dynamic = param[indexNum(2,3)];


        deriv[indexNum(0,2)] = (-x + c) / taux;
        deriv[indexNum(1,2)] = (-theta + c) /tauT;
    }"""

class CL_ReducedSetFitzHughNagumo( ReducedSetFitzHughNagumo, CLModel):
    n_states = 4
    """tau b K11 K12 K21 are nomral variables, f_i IE_i II_i m_i n_i are (1,3) vector, Aik Bik Cik are (3,3) matrices """
    _opencl_ordered_params = "tau b K11 K12 K21 e_i f_i IE_i II_i m_i n_i  Aik Bik Cik ".split()
    _opencl_program_source_file = "ReducedSetFitzHughNagumo.cl"
    _opencl_program_source = """//ReducedSetFitzHugnNagumo
    #define indexNum(cur,totalN) cur*n+i
    #define getFloat3Vector(ptr) (float3)(*ptr,*(ptr+1),*(ptr+2))
    __kernel void dfun(__global float *state, __global float *coupling,
                   __global float *param, __global float *deriv)
    {
        int i = get_global_id(0), n = get_global_size(0);

        float3 xi = vload3(i,state), eta=vload3(i,state+3*n),
        alpha = vload3(i,state+6*n), beta = vload3(i,state+9*n);
        // tau b K11 K12 K21    e_i f_i IE_i II_i m_i n_i (1*3 vector)   Aik Bik Cik (3*3 matrix) params length = 50


        float c_0 = coupling[i];

        float tau = param[i];
        float b = param[n+i];
        float K11 = param[2*n+i];
        float K12 = param[3*n+i];
        float K21 = param[4*n+i];

        float3 e_i = vload3(i,param + 5*n);
        float3 f_i = vload3(i,param + 8*n);
        float3 IE_i = vload3(i,param + 11*n);
        float3 II_i = vload3(i,param +14*n);
        float3 m_i = vload3(i,param+ 17*n);
        float3 n_i = vload3(i,param + 20*n);

        float3 Aik_0 = vload3(i,param + 23*n);
        float3 Aik_1 = vload3(i,param + 23*n+3);
        float3 Aik_2 = vload3(i,param + 23*n+6);

        float3 Bik_0 = vload3(i,param + 32*n);
        float3 Bik_1 = vload3(i,param + 32*n+3);
        float3 Bik_2 = vload3(i,param + 32*n+6);

        float3 Cik_0 = vload3(i,param + 41*n);
        float3 Cik_1 = vload3(i,param + 41*n+3);
        float3 Cik_2 = vload3(i,param + 41*n+6);


        float local_coupling = 0;
        float3 deriv1 = tau * (xi - e_i * pow(xi,3)-eta)+
                                 K11 * ( (float3)(dot(xi,Aik_0),dot(xi,Aik_1),dot(xi,Aik_2)) - xi)-
                                 K12*( (float)(dot(alpha,Bik_0),dot(alpha,Bik_1),dot(alpha,Bik_2) )-xi)+
                                 tau * (IE_i+c_0+local_coupling*xi);

        float3 deriv2 = (xi - b*eta + m_i)/tau;
        float3 deriv3 = tau * (alpha-f_i*pow(alpha,3)/3 - beta)+
                                K21 * ((float3)(dot(xi,Cik_0),dot(xi,Cik_1),dot(xi,Cik_2)) -alpha)+
                                tau * (II_i+c_0+local_coupling*xi);
        float3 deriv4 = (alpha-b*beta+n_i)/tau;

        vstore3(deriv1,i,deriv);
        vstore3(deriv2,i,deriv+3*n);
        vstore3(deriv3,i,deriv+6*n);
        vstore3(deriv4,i,deriv+9*n);

    }

    """




class CL_ReducedSetHindmarshRose( ReducedSetHindmarshRose , CLModel):
    n_states = 6
    _opencl_ordered_params = "r s K11 K12 K21 a_i b_i c_i d_i e_i f_i h_i p_i IE_i II_i m_i n_i   A_ik B_ik C_ik ".split()
    _opencl_program_source_file = "ReducedSetHindmarshRose.cl"
    _opencl_program_source = """

    __kernel void dfun(__global float *state, __global float *coupling,
                   __global float *params, __global float *deriv)
    {
        int i = get_global_id(0), n = get_global_size(0);

        float3 xi = vload3(i,state), eta=vload3(i,state+3*n),
        tau = vload3(i,state+6*n), alpha = vload3(i,state+9*n),
        beta = vload3(i,state+12*n), gamma = vload3(i,state+15*n);

        // "r s K11 K12 K21   a_i b_i c_i d_i e_i f_i h_i p_i IE_i II_i m_i n_i   A_ik B_ik C_ik params length = 68

        float c_0 = coupling[i];

        float r = params[i];
        float s = params[n+i];
        float K11 = params[2*n+i];
        float K12 = params[3*n+i];
        float K21 = params[4*n+i];

        float3 a_i = vload3(i,params + 5*n);
        float3 b_i = vload3(i,params + 8*n);
        float3 c_i = vload3(i,params + 11*n);
        float3 d_i = vload3(i,params + 14*n);
        float3 e_i = vload3(i,params + 17*n);
        float3 f_i = vload3(i,params + 20*n);
        float3 h_i = vload3(i,params + 23*n);
        float3 p_i = vload3(i,params + 26*n);
        float3 IE_i = vload3(i,params + 29*n);
        float3 II_i = vload3(i,params + 32*n);
        float3 m_i = vload3(i,params + 35*n);
        float3 n_i = vload3(i,params + 38*n);



        float3 Aik_0 = vload3(i,params + 41*n);
        float3 Aik_1 = vload3(i,params + 41*n+3);
        float3 Aik_2 = vload3(i,params + 41*n+6);

        float3 Bik_0 = vload3(i,params + 50*n);
        float3 Bik_1 = vload3(i,params + 50*n+3);
        float3 Bik_2 = vload3(i,params + 50*n+6);

        float3 Cik_0 = vload3(i,params + 59*n);
        float3 Cik_1 = vload3(i,params + 59*n+3);
        float3 Cik_2 = vload3(i,params + 59*n+6);

        float local_coupling = 0;
        //TODO Dot Product
        float3 deriv1 = (eta - a_i * pow(xi , 3) + b_i * pow(xi, 2) - tau +
                                K11 * ( (float3)(dot(xi , Aik_0),dot(xi , Aik_1),dot(xi , Aik_2)) - xi) -
                                K12 * ( (float3)(dot(alpha , Bik_0),dot(alpha , Bik_1),dot(alpha , Bik_2) )- xi) +
                                IE_i + c_0 + local_coupling * xi);

        float3 deriv2 = c_i - d_i * pow(xi, 2) - eta;

        float3 deriv3 = r * s * xi - r * tau - m_i;

        float3 deriv4 = beta - e_i * pow(alpha, 3) + f_i * pow(alpha , 2) - gamma +
                                K21 * ( (float3)(dot(xi , Cik_0),dot(xi , Cik_1), dot(xi , Cik_2)) - alpha) +
                                II_i + c_0 + local_coupling * xi;

        float3 deriv5 = h_i - p_i * pow(alpha, 2) - beta;

        float3 deriv6 = r * s * alpha - r * gamma - n_i;


        vstore3(deriv1,i,deriv);
        vstore3(deriv2,i,deriv+3*n);
        vstore3(deriv3,i,deriv+6*n);
        vstore3(deriv4,i,deriv+9*n);
        vstore3(deriv5,i,deriv+12*n);
        vstore3(deriv6,i,deriv+15*n);



    }
    """


class CL_Epileptor( Epileptor, CLModel):
    n_states = 6
    _opencl_ordered_params = "x0 Iext Iext2 a b slope tt Kvf c d r Ks Kf aa tau".split()
    _opencl_program_source_file = "Epileptor.cl"
    _opencl_program_source = """
   #define indexNum(cur,totalN) cur*n+i
__kernel void dfun(__global float *state, __global float *coupling,
                   __global float *param, __global float *deriv)
    {
        int i = get_global_id(0), n = get_global_size(0);

        float y0 = state[indexNum(0,6)], y1=param[indexNum(1,6)],
        y2 = state[indexNum(2,6)], y3 = param[indexNum(4,6)],
        y4 = state[indexNum(4,6)], y5 = param[indexNum(5,6)];

        // x0 Iext Iext2 a b slope tt Kvf c d r Ks Kf aa tau ydot

        float c_pop1 = coupling[indexNum(0,2)];
        float c_pop2 = coupling[indexNum(1,2)];

        float x0 = param[indexNum(0,15)];
        float Iext = param[indexNum(1,15)];
        float Iext2 = param[indexNum(2,15)];
        float a = param[indexNum(3,15)];
        float b = param[indexNum(4,15)];
        float slope = param[indexNum(5,15)];
        float tt = param[indexNum(6,15)];
        float Kvf = param[indexNum(7,15)];
        float c = param[indexNum(8,15)];
        float d = param[indexNum(9,15)];
        float r = param[indexNum(10,15)];
        float Ks = param[indexNum(11,15)];
        float Kf = param[indexNum(12,15)];
        float aa = param[indexNum(13,15)];
        float tau = param[indexNum(14,15)];


        float temp_ydot0,temp_ydot2,temp_ydot4;
        if(y0 < 0.0){
            temp_ydot0 = -a*y0*y0+b*y0;
        }else{
            temp_ydot0 = slope-y3+0.6*(y2-4.0)*(y2-4.0);
        }

        deriv[0] = tt * (y1-y2+Iext + Kvf + c_pop1+temp_ydot0*y0);
        deriv[1] = tt * (c - d*y0*y0 - y1);

        if( y2 < 0.0){
            temp_ydot2 = -0.1*pow(y2,7);
        }else{
            temp_ydot2 = 0.0;
        }
        deriv[2] = tt* (r * (4*(y0-x0)-y2+temp_ydot2+Ks*c_pop1));

        deriv[3] = tt * (-y4 + y3 - pow(y3,3) + Iext2 + 2 * y5 - 0.3 * (y2 - 3.5) + Kf * c_pop2);

        if(y3<-0.25){
            temp_ydot4 = 0.0;
        }else{
            temp_ydot4 = aa*(y3+0.25);
        }
        deriv[4] = tt * ((-y4 + temp_ydot4) / tau);
        deriv[5] = tt * (-0.01 * (y5 - 0.1 * y0));


    }
    """

class CL_Later_Breakspear( LarterBreakspear, CLModel):
    n_states = 3
    _opencl_ordered_params = " gCa gK gL phi gNa TK TCa TNa VCa VK VL VNa d_K tau_K d_Na d_Ca aei aie b C ane ani aee Iext rNMDA VT d_V ZT d_Z QV_max QZ_max t_scale".split()
    _opencl_program_source_file = "Larter_breakspear.cl"
    _opencl_program_source = """
    #define indexNum(cur,totalN) cur*n+i
__kernel void dfun(__global float *state, __global float *coupling,
                   __global float *param, __global float *deriv)
    {
    int i = get_global_id(0), n = get_global_size(0);
    float V = state[indexNum(0,3)], W = param[indexNum(1,3)],Z = param[indexNum(2,3)];

    float c_0 = coupling[i];

    float gCa = param[indexNum(0,32)];
    float gK = param[indexNum(1,32)];
    float gL = param[indexNum(2,32)];
    float phi = param[indexNum(3,32)];
    float gNa = param[indexNum(4,32)];
    float TK = param[indexNum(5,32)];
    float TCa = param[indexNum(6,32)];
    float TNa = param[indexNum(7,32)];
    float VCa = param[indexNum(8,32)];
    float VK = param[indexNum(9,32)];
    float VL = param[indexNum(10,32)];
    float VNa = param[indexNum(11,32)];
    float d_K = param[indexNum(12,32)];
    float tau_K = param[indexNum(13,32)];
    float d_Na = param[indexNum(14,32)];
    float d_Ca = param[indexNum(15,32)];
    float aei = param[indexNum(16,32)];
    float aie = param[indexNum(17,32)];
    float b = param[indexNum(18,32)];
    float C = param[indexNum(19,32)];
    float ane = param[indexNum(20,32)];
    float ani = param[indexNum(21,32)];
    float aee = param[indexNum(22,32)];
    float Iext = param[indexNum(23,32)];
    float rNMDA = param[indexNum(24,32)];
    float VT = param[indexNum(25,32)];
    float d_V = param[indexNum(26,32)];
    float ZT = param[indexNum(27,32)];
    float d_Z = param[indexNum(28,32)];
    float QV_max = param[indexNum(29,32)];
    float QZ_max = param[indexNum(30,32)];
    float t_scale = param[indexNum(31,32)];

    float local_coupling = 1;
    float m_Ca = 0.5 * (1 + tan((V - TCa) / d_Ca));
    float m_Na = 0.5 * (1 + tan((V - TNa) / d_Na));
    float m_K  = 0.5 * (1 + tan((V - TK )  / d_K));
    // voltage to firing rate
    float QV    = 0.5 * QV_max * (1 + tan((V - VT) / d_V));
    float QZ    = 0.5 * QZ_max * (1 + tan((Z - ZT) / d_Z));
    float lc_0  = local_coupling * QV;
    deriv[0] = t_scale * (- (gCa + (1.0 - C) * (rNMDA * aee) * (QV + lc_0)+ C * rNMDA * aee * c_0) * m_Ca * (V - VCa)
                     - gK * W * (V - VK)
                     - gL * (V - VL)
                     - (gNa * m_Na + (1.0 - C) * aee * (QV  + lc_0) + C * aee * c_0) * (V - VNa)
                     - aie * Z * QZ
                     + ane * Iext);
    deriv[1] = t_scale * phi * (m_K - W) / tau_K;
    deriv[2] = t_scale * b * (ani * Iext + aei * V * QV);

    }
   """

class CL_WilsonCowan ( WilsonCowan, CLModel ):
    n_states = 2
    _opencl_ordered_params = " c_ee c_ei c_ie c_ii tau_e tau_i a_e b_e c_e a_i b_i c_i r_e r_i k_e k_i P Q theta_e theta_i alpha_e alpha_i".split()
    _opencl_program_source_file = "Wilson_Cowan.cl"
    _opencl_program_source = """
    #define indexNum(cur,totalN) cur*n+i
    __kernel void dfun(__global float *state, __global float *coupling,
                   __global float *param, __global float *deriv)
    {
        int i = get_global_id(0), n = get_global_size(0);

        float E = state[indexNum(0,2)], I = state[indexNum(1,2)];
        //c_ee c_ei c_ie c_ii tau_e tau_i a_e b_e c_e a_i b_i c_i r_e r_i k_e k_i P Q theta_e theta_i alpha_e alpha_i

        float c_0 = coupling[i];

        float c_ee = param[indexNum(0,22)];
        float c_ei = param[indexNum(1,22)];
        float c_ie = param[indexNum(2,22)];
        float c_ii = param[indexNum(3,22)];
        float tau_e = param[indexNum(4,22)];
        float tau_i = param[indexNum(5,22)];
        float a_e = param[indexNum(6,22)];
        float b_e = param[indexNum(7,22)];
        float c_e = param[indexNum(8,22)];
        float a_i = param[indexNum(9,22)];
        float b_i = param[indexNum(10,22)];
        float c_i = param[indexNum(11,22)];
        float r_e = param[indexNum(12,22)];
        float r_i = param[indexNum(13,22)];
        float k_e = param[indexNum(14,22)];
        float k_i = param[indexNum(15,22)];
        float P = param[indexNum(16,22)];
        float Q = param[indexNum(17,22)];
        float theta_e = param[indexNum(18,22)];
        float theta_i = param[indexNum(19,22)];
        float alpha_e = param[indexNum(20,22)];
        float alpha_i = param[indexNum(21,22)];

        //TODO: dummy local_coupling
        float local_coupling = 1;
        float lc_0 = local_coupling * E;
        float lc_1 = local_coupling * I;

        float x_e = alpha_e * (c_ee * E - c_ei * I + P  - theta_e +  c_0 + lc_0 + lc_1);
        float x_i = alpha_i * (c_ie * E - c_ii * I + Q  - theta_i + lc_0 + lc_1);

        float s_e = c_e / (1.0 + exp(-a_e * (x_e - b_e)));
        float s_i = c_i / (1.0 + exp(-a_i * (x_i - b_i)));

        deriv[indexNum(0,2)] = (-E + (k_e - r_e * E) * s_e) / tau_e;
        deriv[indexNum(1,2)] = (-I + (k_i - r_i * I) * s_i) / tau_i;



    }"""



# -*- coding: utf-8 -*-
#
#
#  TheVirtualBrain-Scientific Package. This package holds all simulators, and
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2017, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
Test for tvb.simulator.models module

.. moduleauthor:: Paula Sanz Leon <sanzleon.paula@gmail.com>

"""

from tvb.tests.library.base_testcase import BaseTestCase
from tvb.simulator import models
import numpy


class TestModels(BaseTestCase):
    """
    Define test cases for models:
        - initialise each class
        - check that initial conditions are always in range

    """

    @staticmethod
    def _validate_initialization(model, expected_sv, expected_models=1):

        model.configure()
        dt = 2 ** -4
        history_shape = (1, model._nvar, 1, model.number_of_modes)
        model_ic = model.initial(dt, history_shape)
        assert expected_sv == model._nvar
        assert expected_models == model.number_of_modes

        svr = model.state_variable_range
        sv = model.state_variables
        for i, (lo, hi) in enumerate([svr[sv[i]] for i in range(model._nvar)]):
            for val in model_ic[:, i, :].flatten():
                assert (lo < val < hi)

        state = numpy.zeros((expected_sv, 10, model.number_of_modes))
        obser = model.observe(state)
        assert (len(model.variables_of_interest), 10, model.number_of_modes) == obser.shape
        return state, obser

    def test_wilson_cowan(self):
        """
        Default parameters are taken from figure 4 of [WC_1972]_, pag. 10
        """
        model = models.WilsonCowan()
        self._validate_initialization(model, 2)

    def test_g2d(self):
        """
        Default parameters:

        +---------------------------+
        |  SanzLeonetAl  2013       |
        +--------------+------------+
        |Parameter     |  Value     |
        +==============+============+
        | a            |    - 0.5   |
        +--------------+------------+
        | b            |    -10.0   |
        +--------------+------------+
        | c            |      0.0   |
        +--------------+------------+
        | d            |      0.02  |
        +--------------+------------+
        | I            |      0.0   |
        +--------------+------------+
        |* excitable regime if      |
        |* intrinsic frequency is   |
        |  approx 10 Hz             |
        +---------------------------+


        """
        model = models.Generic2dOscillator()
        state, obser = self._validate_initialization(model, 2)
        numpy.testing.assert_allclose(obser[0], state[0])

    def test_g2d_voi(self):
        model = models.Generic2dOscillator(
            variables_of_interest=('W', 'V - W')
        )
        (V, W), (voi_W, voi_WmV) = self._validate_initialization(model, 2)
        numpy.testing.assert_allclose(voi_W, W)
        numpy.testing.assert_allclose(voi_WmV, W - V)

    def test_jansen_rit(self):
        """
        """
        model = models.JansenRit()
        self._validate_initialization(model, 6)

    def test_sj2d(self):
        """
        """
        model = models.ReducedSetFitzHughNagumo()
        self._validate_initialization(model, 4, 3)

    def test_sj3d(self):
        """
        """
        model = models.ReducedSetHindmarshRose()
        self._validate_initialization(model, 6, 3)

    def test_reduced_wong_wang(self):
        """
        """
        model = models.ReducedWongWang()
        self._validate_initialization(model, 1)

    def test_zetterberg_jansen(self):
        """
        """
        model = models.ZetterbergJansen()
        self._validate_initialization(model, 12)

    def test_epileptor(self):
        """
        """
        model = models.Epileptor()
        self._validate_initialization(model, 6)

    def test_hopfield(self):
        """
        """
        model = models.Hopfield()
        self._validate_initialization(model, 2)

    def test_kuramoto(self):
        """
        """
        model = models.Kuramoto()
        self._validate_initialization(model, 1)

    def test_larter(self):
        """
        """
        model = models.LarterBreakspear()
        self._validate_initialization(model, 3)

    def test_linear(self):
        model = models.Linear()
        self._validate_initialization(model, 1)

    def test_ww(self):
        model = models.ReducedWongWang()
        self._validate_initialization(model, 1)

        model = models.ReducedWongWangExcInh()
        self._validate_initialization(model, 2)



# -*- coding: utf-8 -*-
#
#
#  TheVirtualBrain-Scientific Package. This package holds all simulators, and 
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2017, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
The TimeSeries datatypes. This brings together the scientific and framework 
methods that are associated with the time-series data.

.. moduleauthor:: Stuart A. Knock <Stuart@tvb.invalid>

"""

from tvb.datatypes import sensors, surfaces, volumes, region_mapping, connectivity
from tvb.basic.logger.builder import get_logger
from tvb.basic.neotraits.api import HasTraits, Attr, NArray, List, Float, narray_summary_info

LOG = get_logger(__name__)


class TimeSeries(HasTraits):
    """
    Base time-series dataType.
    """
    title = Attr(str)

    data = NArray(
        label="Time-series data",
        doc="""An array of time-series data, with a shape of [tpts, :], where ':' represents 1 or more dimensions""")

    labels_ordering = List(
        default=("Time", "State Variable", "Space", "Mode"),
        label="Dimension Names",
        doc="""List of strings representing names of each data dimension""")

    labels_dimensions = Attr(
        field_type=dict,
        default={},
        label="Specific labels for each dimension for the data stored in this timeseries.",
        doc=""" A dictionary containing mappings of the form {'dimension_name' : [labels for this dimension] }""")

    time = NArray(
        label="Time-series time",
        required=False,
        doc="""An array of time values for the time-series, with a shape of [tpts,].
            This is 'time' as returned by the simulator's monitors.""")

    start_time = Float(label="Start Time:")

    sample_period = Float(label="Sample period", default=1.0)

    # Specify the measure unit for sample period (e.g sec, msec, usec, ...)
    sample_period_unit = Attr(
        field_type=str,
        label="Sample Period Measure Unit",
        default="ms")

    @property
    def nr_dimensions(self):
        return self.data.ndim

    @property
    def sample_rate(self):
        return 1.0 / self.sample_period

    def summary_info(self):
        """
        Gather scientifically interesting summary information from an instance of this datatype.
        """
        summary = {
            "Time-series type": self.__class__.__name__,
            "Time-series name": self.title,
            "Dimensions": self.labels_ordering,
            "Time units": self.sample_period_unit,
            "Sample period": self.sample_period,
            "Length": self.sample_period * self.data.shape[0]
        }
        summary.update(narray_summary_info(self.data))
        return summary


class SensorsTSBase(TimeSeries):

    def summary_info(self):
        """
        Gather scientifically interesting summary information from an instance of this datatype.
        """
        summary = super(SensorsTSBase, self).summary_info()
        summary.update({"Source Sensors": self.sensors.title})
        return summary


class TimeSeriesEEG(SensorsTSBase):
    """ A time series associated with a set of EEG sensors. """

    sensors = Attr(field_type=sensors.SensorsEEG)
    labels_ordering = List(of=str, default=("Time", "1", "EEG Sensor", "1"))


class TimeSeriesMEG(SensorsTSBase):
    """ A time series associated with a set of MEG sensors. """

    sensors = Attr(field_type=sensors.SensorsMEG)
    labels_ordering = List(of=str, default=("Time", "1", "MEG Sensor", "1"))


class TimeSeriesSEEG(SensorsTSBase):
    """ A time series associated with a set of Internal sensors. """

    sensors = Attr(field_type=sensors.SensorsInternal)
    labels_ordering = List(of=str, default=("Time", "1", "sEEG Sensor", "1"))


class TimeSeriesRegion(TimeSeries):
    """ A time-series associated with the regions of a connectivity. """

    connectivity = Attr(field_type=connectivity.Connectivity)
    region_mapping_volume = Attr(field_type=region_mapping.RegionVolumeMapping, required=False)
    region_mapping = Attr(field_type=region_mapping.RegionMapping, required=False)
    labels_ordering = List(of=str, default=("Time", "State Variable", "Region", "Mode"))

    def summary_info(self):
        """
        Gather scientifically interesting summary information from an instance of this datatype.
        """
        summary = super(TimeSeriesRegion, self).summary_info()
        summary.update({
            "Source Connectivity": self.connectivity.title,
            "Region Mapping": self.region_mapping.title if self.region_mapping else "None",
            "Region Mapping Volume": (self.region_mapping_volume.title
                                      if self.region_mapping_volume else "None")
        })
        return summary


class TimeSeriesSurface(TimeSeries):
    """ A time-series associated with a Surface. """

    surface = Attr(field_type=surfaces.CorticalSurface)
    labels_ordering = List(of=str, default=("Time", "State Variable", "Vertex", "Mode"))

    def summary_info(self):
        """
        Gather scientifically interesting summary information from an instance of this datatype.
        """
        summary = super(TimeSeriesSurface, self).summary_info()
        summary.update({"Source Surface": self.surface.title})
        return summary


class TimeSeriesVolume(TimeSeries):
    """ A time-series associated with a Volume. """

    volume = Attr(field_type=volumes.Volume)
    labels_ordering = List(of=str, default=("Time", "X", "Y", "Z"))

    def summary_info(self):
        """
        Gather scientifically interesting summary information from an instance of this datatype.
        """
        summary = super(TimeSeriesVolume, self).summary_info()
        summary.update({"Source Volume": self.volume.title})
        return summary



# -*- coding: utf-8 -*-
#
#
#  TheVirtualBrain-Scientific Package. This package holds all simulators, and
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2017, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

import pyopencl
import pyopencl.array
import numpy
from .models import CLComponent, CLModel
from ..integrators import HeunDeterministic, HeunStochastic, EulerDeterministic, EulerStochastic, RungeKutta4thOrderDeterministic , Identity
from .util import *

grw = lambda arr, ctx: pyopencl.Buffer(ctx, pyopencl.mem_flags.COPY_HOST_PTR, hostbuf=arr)

class CLIntegrator(CLComponent):
    _opencl_program_source_file = "baseIntegrator.cl"
    # vectors = lenOfVec * size, numOfThreads = size = numOfCoefficients
    _opencl_program_source = """
    __kernel void integrate( int    size   , int lenOfVec,
                            __global float *vectors, __global float *coefficients,__global float *ans
        ){
    int id = get_global_id(0), n = get_global_size(0);
    //printf("vectors: %d ans: %d ",vectors[id], ans[id]);

    for(int i = 0; i < size; i++){
        ans[id]+=vectors[i*lenOfVec+id]*coefficients[i];
    }
    }"""

    def cl_integrate(self, numOfCoefficient, lenOfVec, vectors, coefficients, ):
        n_states = vectors.shape[0]
        n_nodes = lenOfVec
        n_mode = vectors.shape[2]

        size = numpy.int32(numOfCoefficient)
        lenOfVec = numpy.int32(lenOfVec)
        vec_cl = grw(vectors, self._context)
        co_cl = grw(coefficients, self._context)
        ans = numpy.zeros((n_states,n_nodes,n_mode))
        ans_cl = grw(ans, self._context)
        args = size,lenOfVec,vec_cl,co_cl,ans_cl
        self._kernel(self._queue, (int(lenOfVec),), None, *args)


        pyopencl.enqueue_copy( self._queue, ans , ans_cl)

        return ans

    def _alloc_opencl(self, n_nodes, n_states=1, n_mode=1 , stochasticVariant = False):
        if not hasattr(self, '_queue'):
            msg = "OpenCL components must be configured via the `configure_opencl` method prior to use."
            raise RuntimeError(msg)

        arrays = {'state': (n_states, n_nodes, n_mode), 'coupling': (1, n_nodes), 'deriv': (n_states, n_nodes, n_mode)}
        self._arrays = {}
        if stochasticVariant:
            arrays['noise'] = (n_states, n_nodes, n_mode)

        for name, shape in list(arrays.items()):
            self._arrays[name] = pyopencl.array.Array(self._queue, shape, 'f')
            #print "Name: type:",isinstance(self._arrays[name],pyopencl.Buffer)
            #print "name: %s, shape %s"%(name,shape)

    def scheme(self,  X , dfun, coupling, local_coupling = 0.0, stimulus = 0.0):
        n_states = X.shape[0]
        n_nodes = X.shape[1]
        n_mode = X.shape[2]
        print("X shape:",X.shape)
        # check for openclArray buffer
        if not hasattr(self, '_arrays'):
            self._alloc_opencl( n_nodes,n_states = n_states, n_mode = n_mode , stochasticVariant = True )

        print("self: %s %s X: %s %s"%(self._arrays['state'].dtype, self._arrays['state'].size, X.dtype, X.size))
        self._arrays['state'].set( X )

        self._arrays['coupling'].set( coupling )

        #X_cl = pyopencl.array.Array(self._queue, X.shape ,'f')
        #coupling_cl = pyopencl.array.Array(self._queue, coupling.shape ,'f')
        X_next = dfun(self._arrays['state'], self._arrays['coupling'])

        return X_next



    def setup_method(self):
        self._context, self._queue = context_and_queue(create_cpu_context())

    def configure_opencl(self, context, queue):
        super(CLIntegrator, self).configure_opencl(context, queue)

        self._kernel = self._program.integrate


class CL_EulerDeterministic( CLIntegrator, EulerDeterministic ):
    #_opencl_program_source_file = "EulerDeterministic.cl"
    # opencl_program_source = ""

    def scheme(self, X, dfun, coupling, local_coupling=0.0, stimulus=0.0):
        n_states = X.shape[0]
        n_nodes = X.shape[1]
        n_mode = X.shape[2]
        print("X shape:", X.shape)
        # check for openclArray buffer
        if not hasattr(self, '_arrays'):
            self._alloc_opencl(n_nodes, n_states=n_states, n_mode=n_mode, stochasticVariant=True)

        #print "self: %s %s X: %s %s" % (self._arrays['state'].dtype, self._arrays['state'].size, X.dtype, X.size)
        self._arrays['state'].set(X)
        self._arrays['coupling'].set(coupling)
        self.dX = dfun(self._arrays['state'], self._arrays['coupling'])
        #X_next = self.dX + stimulus
        X_next = self._arrays['state'] + self.dt * (self.dX + stimulus)
        return X_next


class CL_Identity(CLIntegrator, Identity):

    def scheme(self, X, dfun, coupling=None, local_coupling=0.0, stimulus=0.0):
        n_states = X.shape[0]
        n_nodes = X.shape[1]
        n_mode = X.shape[2]
        print("X shape:", X.shape)
        # check for openclArray buffer
        if not hasattr(self, '_arrays'):
            self._alloc_opencl(n_nodes, n_states=n_states, n_mode=n_mode, stochasticVariant=True)

        self._arrays['state'].set(X)

        self._arrays['coupling'].set(coupling)
        self.dX = dfun(self._arrays['state'], self._arrays['coupling'])

        return self.dX + stimulus



class CL_HeunDeterministic(CLIntegrator, HeunDeterministic):
    def _alloc_opencl(self, n_nodes, n_states=1, n_mode=1):
        super(CL_HeunDeterministic, self)._alloc_opencl(n_nodes, n_states, n_mode, stochasticVariant=True)

        self._arrays['inter'] = pyopencl.array.Array(self._queue, (n_states, n_nodes, n_mode), 'f')

    def scheme(self, X, dfun, coupling= None, local_coupling=0.0, stimulus=0.0):
        n_states = X.shape[0]
        n_nodes = X.shape[1]
        n_mode = X.shape[2]
        print("X shape:", X.shape)
        # check for openclArray buffer
        if not hasattr(self, '_arrays'):
            self._alloc_opencl(n_nodes, n_states=n_states, n_mode=n_mode)

        self._arrays['state'].set(X)

        self._arrays['coupling'].set(coupling)
        self._arrays['inter'] = dfun(self._arrays['state'], self._arrays['coupling'])


        inter = self._arrays['state'] + self.dt * (self._arrays['inter'] + stimulus)
        #print "shceme: m_dx_tn",m_dx_tn.shape ,"inter",inter.shape
        self.clamp_state(inter)

        dX = (self._arrays['inter'] + dfun(inter, self._arrays['coupling'])) * self.dt / 2.0
        #print "scheme dX.shape: X_cl shape self.dt stimulus ",dX.shape, X_cl.shape, self.dt,stimulus
        X_next = self._arrays['state'] + dX + self.dt * stimulus

        self.clamp_state(X_next)

        return X_next


class CL_EulerStochastic(CLIntegrator, EulerStochastic):
    def _alloc_opencl(self, n_nodes, n_states=1, n_mode=1 ):
        super( CL_EulerStochastic, self)._alloc_opencl(n_nodes, n_states, n_mode, stochasticVariant= True )

        self._arrays['inter'] = pyopencl.array.Array(self._queue, (n_states, n_nodes, n_mode), 'f')

    def scheme(self, X, dfun, coupling, local_coupling = 0.0, stimulus = 0.0):
        n_states = X.shape[0]
        n_nodes = X.shape[1]
        n_mode = X.shape[2]
        print("X shape:", X.shape)
        # check for openclArray buffer
        if not hasattr(self, '_arrays'):
            self._alloc_opencl( n_nodes, n_states=n_states, n_mode=n_mode )

        self._arrays['state'].set(X)
        self._arrays['coupling'].set(coupling)
        noise = self.noise.generate(X.shape).astype(numpy.float32)
        self._arrays['noise'].set(noise)

        dX = dfun(self._arrays['state'], self._arrays['coupling']) * self.dt
        noise_gfun = self.noise.gfun(X)
        X_next =  self._arrays['state'] + dX + noise_gfun * self._arrays['noise'] + self.dt * stimulus
        self.clamp_state(X_next)
        return X_next



class CL_HeunStochastic( CLIntegrator, HeunStochastic ):
    def _alloc_opencl(self, n_nodes, n_states=1, n_mode=1):
        super(CL_HeunStochastic, self)._alloc_opencl(n_nodes, n_states, n_mode, stochasticVariant=True)

        self._arrays['inter'] = pyopencl.array.Array(self._queue, (n_states, n_nodes, n_mode), 'f')

    def scheme(self,  X , dfun, coupling, local_coupling = 0.0, stimulus = 0.0):
        n_states = X.shape[0]
        n_nodes = X.shape[1]
        n_mode = X.shape[2]
        print("X shape:", X.shape)
        # check for openclArray buffer
        if not hasattr(self, '_arrays'):
            self._alloc_opencl(n_nodes, n_states=n_states, n_mode=n_mode )

        self._arrays['state'].set(X)
        self._arrays['coupling'].set(coupling)

        noise = self.noise.generate(X.shape).astype(numpy.float32)
        self._arrays['noise'].set(noise)

        noise_gfun = self.noise.gfun(X)

        if (noise_gfun.shape != (1,) and noise.shape[0] != noise_gfun.shape[0]):
            msg = str("Got shape %s for noise but require %s."
                      " You need to reconfigure noise after you have changed your model." % (
                          noise_gfun.shape, (noise.shape[0], noise.shape[1])))
            raise Exception(msg)

        m_dx_tn = dfun(self._arrays['state'], self._arrays['coupling'])

        noise *= noise_gfun

        self._arrays['inter'] = self._arrays['state'] + self.dt * m_dx_tn + self._arrays['noise'] + self.dt * stimulus
        self.clamp_state(self._arrays['inter'])

        dX = (m_dx_tn + dfun(self._arrays['inter'], self._arrays['coupling'] )) * self.dt / 2.0

        X_next = self._arrays['state'] + dX + self._arrays['noise'] + self.dt * stimulus
        self.clamp_state(X_next)

        return X_next




class CL_RungeKutta4thOrderDeterministic( CLIntegrator, RungeKutta4thOrderDeterministic):
    def scheme(self,  X , dfun, coupling, local_coupling = 0.0, stimulus = 0.0):

        n_states = X.shape[0]
        n_nodes = X.shape[1]
        n_mode = X.shape[2]
        print("X shape:", X.shape)
        # check for openclArray buffer
        if not hasattr(self, '_arrays'):
            self._alloc_opencl(n_nodes, n_states=n_states, n_mode=n_mode)

        self._arrays['state'].set(X)
        self._arrays['coupling'].set(coupling)

        dt = self.dt
        dt2 = dt / 2.0
        dt6 = dt / 6.0

        k1 = dfun(self._arrays['state'], self._arrays['coupling'])
        self._arrays['inter'] = self._arrays['state'] + dt2 * k1

        self.clamp_state(self._arrays['inter'])
        k2 = dfun(self._arrays['inter'], self._arrays['coupling'])
        self._arrays['inter'] = self._arrays['state'] + dt2 * k2

        self.clamp_state( self._arrays['inter'])
        k3 = dfun( self._arrays['inter'], self._arrays['coupling'])
        self._arrays['inter'] = self._arrays['state'] + dt * k3
        self.clamp_state(self._arrays['inter'])
        k4 = dfun(self._arrays['inter'], self._arrays['coupling'])

        dX = dt6 * (k1 + 2.0 * k2 + 2.0 * k3 + k4)

        X_next = self._arrays['state'] + dX + self.dt * stimulus
        self.clamp_state(X_next)
        return X_next





# -*- coding: utf-8 -*-
#
#
#  TheVirtualBrain-Scientific Package. This package holds all simulators, and 
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2017, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#
"""
DataType for storing a simulator's state in files and as DB reference.

.. moduleauthor:: Lia Domide <lia.domide@codemart.ro>
"""
from tvb.basic.neotraits.api import HasTraits, NArray, Int


class SimulationState(HasTraits):
    """
    Simulation State, prepared for H5 file storage.
    """

    # History Array
    history = NArray(required=False)

    # State array, all state variables
    current_state = NArray(required=False)

    # Simulator's current step number (in time)
    current_step = Int()

    # Array with _stock array for every monitor configured in current simulation.
    # As the monitors are dynamic, we prepare a bunch of arrays for storage in H5 file.
    monitor_stock_1 = NArray(required=False)
    monitor_stock_2 = NArray(required=False)
    monitor_stock_3 = NArray(required=False)
    monitor_stock_4 = NArray(required=False)
    monitor_stock_5 = NArray(required=False)
    monitor_stock_6 = NArray(required=False)
    monitor_stock_7 = NArray(required=False)
    monitor_stock_8 = NArray(required=False)
    monitor_stock_9 = NArray(required=False)
    monitor_stock_10 = NArray(required=False)
    monitor_stock_11 = NArray(required=False)
    monitor_stock_12 = NArray(required=False)
    monitor_stock_13 = NArray(required=False)
    monitor_stock_14 = NArray(required=False)
    monitor_stock_15 = NArray(required=False)


    def __init__(self, **kwargs):
        """ 
        Constructor for Simulator State
        """
        super(SimulationState, self).__init__(**kwargs)
        self.visible = False


    def populate_from(self, simulator_algorithm):
        """
        Prepare a state for storage from a Simulator object.
        """
        self.history = simulator_algorithm.history.buffer.copy()
        self.current_step = simulator_algorithm.current_step
        self.current_state = simulator_algorithm.current_state

        for i, monitor in enumerate(simulator_algorithm.monitors):
            field_name = "monitor_stock_" + str(i + 1)
            setattr(self, field_name, monitor._stock)

            if hasattr(monitor, "_ui_name"):
                self.set_metadata({'monitor_name': monitor._ui_name}, field_name)
            else:
                self.set_metadata({'monitor_name': monitor.__class__.__name__}, field_name)


    def fill_into(self, simulator_algorithm):
        """
        Populate a Simulator object from current stored-state.
        """
        simulator_algorithm.history.initialize(self.history)
        simulator_algorithm.current_step = self.current_step
        simulator_algorithm.current_state = self.current_state

        for i, monitor in enumerate(simulator_algorithm.monitors):
            monitor._stock = getattr(self, "monitor_stock_" + str(i + 1))
  
    
 


# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and 
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2017, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
# CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
DataTypes for mapping some TVB DataTypes to a Connectivity (regions).

In FreeSurfer terms, a RegionMapping is a parcellation and a VolumeMapping is a segmentation.

.. moduleauthor:: Lia Domide <lia.domide@codemart.ro>
.. moduleauthor:: Mihai Andrei <mihai.andrei@codemart.ro>
"""

import numpy
from tvb.basic.readers import try_get_absolute_path, FileReader
from tvb.datatypes.connectivity import Connectivity
from tvb.datatypes.surfaces import Surface
from tvb.datatypes.volumes import Volume
from tvb.basic.logger.builder import get_logger
from tvb.basic.neotraits.api import HasTraits, Attr, NArray

LOG = get_logger(__name__)


class RegionMapping(HasTraits):
    """
    An array (of length Surface.vertices). Each value is representing the index in Connectivity regions
    to which the current vertex is mapped.
    """

    array_data = NArray(dtype=int)

    connectivity = Attr(field_type=Connectivity)

    surface = Attr(field_type=Surface)

    @staticmethod
    def from_file(source_file="regionMapping_16k_76.txt", instance=None):

        if instance is None:
            result = RegionMapping()
        else:
            result = instance

        source_full_path = try_get_absolute_path("tvb_data.regionMapping", source_file)
        reader = FileReader(source_full_path)

        result.array_data = reader.read_array(dtype=numpy.int32)
        return result

    def get_region_mapping_slice(self, start_idx, end_idx):
        """
        Get a slice of the region mapping as used by the region viewers.
        For each vertex on the surface, alpha-indices will be the closest
        region-index
        :param start_idx: vertex index on the surface
        :param end_idx: vertex index on the surface
        :return: NumPy array with [closest_reg_idx ...]
        """
        if isinstance(start_idx, str):
            start_idx = int(start_idx)
        if isinstance(end_idx, str):
            end_idx = int(end_idx)

        return self.array_data[start_idx: end_idx].T


    def get_triangles_mapping(self):
        """
        :return Numpy array of length triangles and for each the region corresponding to one of its vertices.
        """
        triangles_no = self.surface.number_of_triangles
        result = []
        for i in range(triangles_no):
            result.append(self.array_data[self.surface.triangles[i][0]])
        return numpy.array(result)


class RegionVolumeMapping(HasTraits):
    """
    Each value is representing the index in Connectivity regions to which the current voxel is mapped.
    """

    array_data = NArray(dtype=int)

    connectivity = Attr(Connectivity)

    volume = Attr(Volume)

